from unittest.mock import patch, MagicMock
import csle_common.constants.constants as constants
from csle_common.controllers.sdn_controller_manager import SDNControllerManager
import csle_collector.ryu_manager.ryu_manager_pb2 as ryu_manager_pb2
import csle_collector.ryu_manager.ryu_manager_pb2_grpc as ryu_manager_pb2_grpc


class TestSDNControllerSuite:
    """
    Test SDNControllerManager
    """

    @patch("csle_common.util.emulation_util.EmulationUtil.connect_admin")
    @patch("csle_common.util.emulation_util.EmulationUtil.execute_ssh_cmd")
    @patch("time.sleep", return_value=None)
    def test_start_ryu_manager(self, mock_sleep, mock_execute_ssh_cmd, mock_connect_admin) -> None:
        """
        Test method for starting the Ryu manager

        :param mock_execute_ssh_cmd: mock_execute_ssh_cmd
        :param mock_connect_admin: mock_connect_admin
        :param mock_sleep: mock_sleep
        :return: None
        """
        logger = MagicMock()
        sdn_controller_config = MagicMock(
            container=MagicMock(docker_gw_bridge_ip="192.168.1.2"), manager_port=6633, manager_log_dir="/var/log/ryu",
            manager_log_file="ryu.log", manager_max_workers=4)
        emulation_env_config = MagicMock(sdn_controller_config=sdn_controller_config, name="Test Emulation")
        mock_execute_ssh_cmd.side_effect = [
            (b"", b"", 0),  # output of cp /etc/sudoers.bak /etc/sudoers
            (b"", b"", 0),  # output of adding CVE entry to sudoers
            (b"", b"", 0),  # output of chmod 440 /etc/sudoers
        ]
        SDNControllerManager.start_ryu_manager(emulation_env_config, logger)
        mock_connect_admin.assert_any_call(emulation_env_config=emulation_env_config, ip="192.168.1.2",
                                           create_producer=False)
        mock_execute_ssh_cmd.assert_called()

    @patch("csle_common.util.emulation_util.EmulationUtil.connect_admin")
    @patch("csle_common.util.emulation_util.EmulationUtil.execute_ssh_cmd")
    @patch("time.sleep", return_value=None)
    def test_stop_ryu_manager(self, mock_sleep, mock_execute_ssh_cmd, mock_connect_admin) -> None:
        """
        Test method for stopping the Ryu manager

        :param mock_execute_ssh_cmd: mock_execute_ssh_cmd
        :param mock_connect_admin: mock_connect_admin
        :param mock_sleep: mock_sleep
        :return: None
        """
        logger = MagicMock()
        sdn_controller_config = MagicMock(container=MagicMock(docker_gw_bridge_ip="192.168.1.2"))
        emulation_env_config = MagicMock(sdn_controller_config=sdn_controller_config, name="Test Emulation")
        mock_execute_ssh_cmd.side_effect = [(b"", b"", 0)]  # output of pkill ryu_manager
        SDNControllerManager.stop_ryu_manager(emulation_env_config, logger)
        mock_connect_admin.assert_any_call(emulation_env_config=emulation_env_config, ip="192.168.1.2",
                                           create_producer=False)
        mock_execute_ssh_cmd.assert_called()

    @patch("csle_common.controllers.sdn_controller_manager.SDNControllerManager.start_ryu_manager")
    @patch("csle_common.controllers.sdn_controller_manager.SDNControllerManager.get_ryu_status_by_port_and_ip")
    def test_get_ryu_status(self, mock_get_ryu_status_by_port_and_ip, mock_get_ryu_manager) -> None:
        """
        Test method for querying the RyuManager about the status of the Ryu SDN controller

        :param mock_get_ryu_status_by_port_and_ip: mock_get_ryu_status_by_port_and_ip
        :param mock_get_ryu_manager: mock_get_ryu_manager
        :return: None
        """
        logger = MagicMock()
        ryu_dto_mock = MagicMock(spec=ryu_manager_pb2.RyuDTO)
        mock_get_ryu_status_by_port_and_ip.return_value = ryu_dto_mock
        sdn_controller_config = MagicMock(container=MagicMock(docker_gw_bridge_ip="192.168.1.2"), manager_port=6633)
        emulation_env_config = MagicMock(sdn_controller_config=sdn_controller_config, name="Test Emulation")
        result = SDNControllerManager.get_ryu_status(emulation_env_config=emulation_env_config, logger=logger)
        assert result == ryu_dto_mock
        mock_get_ryu_manager.assert_called()
        mock_get_ryu_status_by_port_and_ip.assert_called()

    @patch("grpc.insecure_channel")
    @patch("csle_collector.ryu_manager.query_ryu_manager.get_ryu_status")
    def test_get_ryu_status_by_port_and_ip(self, mock_get_ryu_status, mock_insecure_channel) -> None:
        """
        Test method for querying the RyuManager about the status of the Ryu SDN controller

        :param mock_get_ryu_status: mock_get_ryu_status
        :param mock_insecure_channel: mock_insecure_channel
        :return: None
        """
        ryu_dto_mock = MagicMock(spec=ryu_manager_pb2.RyuDTO)
        mock_get_ryu_status.return_value = ryu_dto_mock
        channel_mock = MagicMock()
        mock_insecure_channel.return_value.__enter__.return_value = channel_mock
        ip = "192.168.1.2"
        port = 6633
        result = SDNControllerManager.get_ryu_status_by_port_and_ip(ip, port)
        mock_insecure_channel.assert_called_once_with(f"{ip}:{port}", options=constants.GRPC_SERVERS.GRPC_OPTIONS)
        mock_get_ryu_status.assert_called_once()
        assert isinstance(mock_get_ryu_status.call_args[0][0], ryu_manager_pb2_grpc.RyuManagerStub)
        assert result == ryu_dto_mock

    @patch("csle_common.controllers.sdn_controller_manager.SDNControllerManager.start_ryu_manager")
    @patch("grpc.insecure_channel")
    @patch("csle_collector.ryu_manager.query_ryu_manager.stop_ryu")
    @patch("time.sleep", return_value=None)
    def test_stop_ryu(self, mock_sleep, mock_stop_ryu, mock_insecure_channel, mock_start_manager) -> None:
        """
        Test method for requesting the RyuManager to stop the RYU SDN controller

        :param mock_stop_ryu: mock_stop_ryu
        :param mock_insecure_channel: mock_insecure_channel
        :param mock_start_manager: mock_start_manager
        :param mock_sleep: mock_sleep
        :return: None
        """
        logger = MagicMock()
        ryu_dto_mock = MagicMock(spec=ryu_manager_pb2.RyuDTO)
        mock_stop_ryu.return_value = ryu_dto_mock
        channel_mock = MagicMock()
        mock_insecure_channel.return_value.__enter__.return_value = channel_mock
        sdn_controller_config = MagicMock(container=MagicMock(docker_gw_bridge_ip="192.168.1.2"), manager_port=6633)
        emulation_env_config = MagicMock(sdn_controller_config=sdn_controller_config, name="Test Emulation")
        ip = "192.168.1.2"
        port = 6633
        result = SDNControllerManager.stop_ryu(emulation_env_config=emulation_env_config, logger=logger)
        mock_insecure_channel.assert_called_once_with(f"{ip}:{port}", options=constants.GRPC_SERVERS.GRPC_OPTIONS)
        mock_stop_ryu.assert_called_once()
        assert isinstance(mock_stop_ryu.call_args[0][0], ryu_manager_pb2_grpc.RyuManagerStub)
        assert result == ryu_dto_mock

    @patch("csle_common.controllers.sdn_controller_manager.SDNControllerManager.start_ryu_manager")
    @patch("grpc.insecure_channel")
    @patch("csle_collector.ryu_manager.query_ryu_manager.start_ryu")
    @patch("time.sleep", return_value=None)
    def test_start_ryu(self, mock_sleep, mock_start_ryu, mock_insecure_channel, mock_start_manager) -> None:
        """
        Test method for requesting the RyuManager to start the RYU SDN controller

        :param mock_start_ryu: mock_start_ryu
        :param mock_insecure_channel: mock_insecure_channel
        :param mock_start_manager: mock_start_manager
        :param mock_sleep: mock_sleep
        :return: None
        """
        logger = MagicMock()
        ryu_dto_mock = MagicMock(spec=ryu_manager_pb2.RyuDTO)
        mock_start_ryu.return_value = ryu_dto_mock
        channel_mock = MagicMock()
        mock_insecure_channel.return_value.__enter__.return_value = channel_mock
        sdn_controller_config = MagicMock(
            container=MagicMock(docker_gw_bridge_ip="192.168.1.2", physical_host_ip="192.168.1.1"),
            manager_port=6633,
            controller_port=6653,
            controller_web_api_port=8080,
            controller_module_name="controller_module",
        )
        emulation_env_config = MagicMock(sdn_controller_config=sdn_controller_config, name="Test Emulation")
        ip = "192.168.1.1"

        result = SDNControllerManager.start_ryu(emulation_env_config=emulation_env_config, physical_server_ip=ip,
                                                logger=logger)
        mock_insecure_channel.assert_called_once_with("192.168.1.2:6633", options=constants.GRPC_SERVERS.GRPC_OPTIONS)
        mock_start_ryu.assert_called_once()
        assert isinstance(mock_start_ryu.call_args[0][0], ryu_manager_pb2_grpc.RyuManagerStub)
        assert result == ryu_dto_mock

    @patch("csle_common.controllers.sdn_controller_manager.SDNControllerManager.start_ryu_manager")
    @patch("grpc.insecure_channel")
    @patch("csle_collector.ryu_manager.query_ryu_manager.start_ryu_monitor")
    def test_start_ryu_monitor(self, mock_start_ryu_monitor, mock_insecure_channel, mock_start_manager) -> None:
        """
        Test method for requesting the RyuManager to start the Ryu monitor

        :param mock_start_ryu_monitor: mock_start_ryu_monitor
        :param mock_insecure_channel: mock_insecure_channel
        :param mock_start_manager: mock_start_manager
        :return: None
        """
        logger = MagicMock()
        ryu_dto_mock = MagicMock(spec=ryu_manager_pb2.RyuDTO)
        mock_start_ryu_monitor.return_value = ryu_dto_mock
        channel_mock = MagicMock()
        mock_insecure_channel.return_value.__enter__.return_value = channel_mock
        sdn_controller_config = MagicMock(
            container=MagicMock(docker_gw_bridge_ip="192.168.1.2", physical_host_ip="192.168.1.1"),
            manager_port=6633, time_step_len_seconds=10)
        kafka_config = MagicMock(container=MagicMock(get_ips=MagicMock(return_value=["192.168.1.3"])), kafka_port=9092)
        emulation_env_config = MagicMock(sdn_controller_config=sdn_controller_config, kafka_config=kafka_config,
                                         name="Test Emulation")
        ip = "192.168.1.1"
        result = SDNControllerManager.start_ryu_monitor(emulation_env_config=emulation_env_config,
                                                        physical_server_ip=ip, logger=logger)
        mock_insecure_channel.assert_called_once_with("192.168.1.2:6633", options=constants.GRPC_SERVERS.GRPC_OPTIONS)
        mock_start_ryu_monitor.assert_called_once()
        assert isinstance(mock_start_ryu_monitor.call_args[0][0], ryu_manager_pb2_grpc.RyuManagerStub)
        assert result == ryu_dto_mock

    @patch("csle_common.controllers.sdn_controller_manager.SDNControllerManager.start_ryu_manager")
    @patch("grpc.insecure_channel")
    @patch("csle_collector.ryu_manager.query_ryu_manager.stop_ryu_monitor")
    def test_stop_ryu_monitor(self, mock_stop_ryu_monitor, mock_insecure_channel, mock_start_manager) -> None:
        """
        Test method for requesting the RyuManager to stop the Ryu monitor

        :param mock_stop_ryu_monitor: mock_stop_ryu_monitor
        :param mock_insecure_channel: mock_insecure_channel
        :param mock_start_manager: mock_start_manager
        :return: None
        """
        logger = MagicMock()
        ryu_dto_mock = MagicMock(spec=ryu_manager_pb2.RyuDTO)
        mock_stop_ryu_monitor.return_value = ryu_dto_mock
        channel_mock = MagicMock()
        mock_insecure_channel.return_value.__enter__.return_value = channel_mock
        sdn_controller_config = MagicMock(
            container=MagicMock(docker_gw_bridge_ip="192.168.1.2", physical_host_ip="192.168.1.1"), manager_port=6633)
        emulation_env_config = MagicMock(sdn_controller_config=sdn_controller_config, name="Test Emulation")
        result = SDNControllerManager.stop_ryu_monitor(emulation_env_config=emulation_env_config, logger=logger)
        mock_insecure_channel.assert_called_once_with("192.168.1.2:6633", options=constants.GRPC_SERVERS.GRPC_OPTIONS)
        mock_stop_ryu_monitor.assert_called_once()
        assert isinstance(mock_stop_ryu_monitor.call_args[0][0], ryu_manager_pb2_grpc.RyuManagerStub)
        assert result == ryu_dto_mock

    def test_get_ryu_managers_ips(self) -> None:
        """
        Test method that extracts the IPS of the Ryu managers in a given emulation

        :return: None
        """
        sdn_controller_config = MagicMock(container=MagicMock(docker_gw_bridge_ip="192.168.1.2"))
        emulation_env_config = MagicMock(sdn_controller_config=sdn_controller_config, name="Test Emulation")
        result = SDNControllerManager.get_ryu_managers_ips(emulation_env_config)
        assert result == ["192.168.1.2"]

    def test_get_ryu_managers_ports(self) -> None:
        """
        Test method that extracts the ports of the Ryu managers in a given emulation

        :return: None
        """
        sdn_controller_config = MagicMock(manager_port=6633)
        emulation_env_config = MagicMock(sdn_controller_config=sdn_controller_config, name="Test Emulation")
        result = SDNControllerManager.get_ryu_managers_ports(emulation_env_config)
        assert result == [6633]

    @patch("csle_common.controllers.sdn_controller_manager.SDNControllerManager.get_ryu_managers_ips")
    @patch("csle_common.controllers.sdn_controller_manager.SDNControllerManager.get_ryu_managers_ports")
    @patch("csle_common.controllers.sdn_controller_manager.SDNControllerManager.get_ryu_status_by_port_and_ip")
    @patch("csle_common.util.emulation_util.EmulationUtil.physical_ip_match")
    @patch("csle_collector.ryu_manager.ryu_manager_util.RyuManagerUtil.ryu_dto_empty", return_value=MagicMock())
    def test_get_ryu_managers_info(self, mock_ryu_dto_empty, mock_physical_ip_match, mock_get_ryu_status_by_port_and_ip,
                                   mock_get_ryu_managers_ports, mock_get_ryu_managers_ips) -> None:
        """
        Test method that extracts the information of the Ryu managers for a given emulation

        :param mock_ryu_dto_empty: mock_ryu_dto_empty
        :param mock_physical_ip_match: mock_physical_ip_match
        :param mock_get_ryu_status_by_port_and_ip: mock_get_ryu_status_by_port_and_ip
        :param mock_get_ryu_managers_ports: mock_get_ryu_managers_ports
        :param mock_get_ryu_managers_ips: mock_get_ryu_managers_ips
        :return: None
        """
        logger = MagicMock()
        sdn_controller_config = MagicMock(
            container=MagicMock(docker_gw_bridge_ip="192.168.1.2", physical_host_ip="192.168.1.1"), manager_port=6633)
        emulation_env_config = MagicMock(sdn_controller_config=sdn_controller_config, execution_id="123",
                                         name="Test Emulation")
        mock_get_ryu_managers_ips.return_value = ["192.168.1.2"]
        mock_get_ryu_managers_ports.return_value = [6633]
        mock_physical_ip_match.return_value = True
        mock_get_ryu_status_by_port_and_ip.return_value = MagicMock()

        active_ips = ["192.168.1.2"]
        physical_server_ip = "192.168.1.1"

        result = SDNControllerManager.get_ryu_managers_info(emulation_env_config, active_ips, logger,
                                                            physical_server_ip)
        mock_get_ryu_managers_ips.assert_called_once_with(emulation_env_config=emulation_env_config)
        mock_get_ryu_managers_ports.assert_called_once_with(emulation_env_config=emulation_env_config)
        mock_physical_ip_match.assert_called_once_with(emulation_env_config=emulation_env_config,
                                                       ip="192.168.1.2", physical_host_ip="192.168.1.1")
        mock_get_ryu_status_by_port_and_ip.assert_called_once_with(port=6633, ip="192.168.1.2")
        assert result



from enum import IntEnum


class EmulationAttackerActionId(IntEnum):
    """
    Enum representing the different attacks in the emulation
    """
    TCP_SYN_STEALTH_SCAN_HOST = 0
    PING_SCAN_HOST = 1
    UDP_PORT_SCAN_HOST = 2
    TCP_CON_NON_STEALTH_SCAN_HOST = 3
    TCP_FIN_SCAN_HOST = 4
    TCP_NULL_SCAN_HOST = 5
    TCP_XMAS_TREE_SCAN_HOST = 6
    OS_DETECTION_SCAN_HOST = 7
    VULSCAN_HOST = 8
    NMAP_VULNERS_HOST = 9
    TELNET_SAME_USER_PASS_DICTIONARY_HOST = 10
    SSH_SAME_USER_PASS_DICTIONARY_HOST = 11
    FTP_SAME_USER_PASS_DICTIONARY_HOST = 12
    CASSANDRA_SAME_USER_PASS_DICTIONARY_HOST = 13
    IRC_SAME_USER_PASS_DICTIONARY_HOST = 14
    MONGO_SAME_USER_PASS_DICTIONARY_HOST = 15
    MYSQL_SAME_USER_PASS_DICTIONARY_HOST = 16
    SMTP_SAME_USER_PASS_DICTIONARY_HOST = 17
    POSTGRES_SAME_USER_PASS_DICTIONARY_HOST = 18
    TCP_SYN_STEALTH_SCAN_ALL = 19
    PING_SCAN_ALL = 20
    UDP_PORT_SCAN_ALL = 21
    TCP_CON_NON_STEALTH_SCAN_ALL = 22
    TCP_FIN_SCAN_ALL = 23
    TCP_NULL_SCAN_ALL = 24
    TCP_XMAS_TREE_SCAN_ALL = 25
    OS_DETECTION_SCAN_ALL = 26
    VULSCAN_ALL = 27
    NMAP_VULNERS_ALL = 28
    TELNET_SAME_USER_PASS_DICTIONARY_ALL = 29
    SSH_SAME_USER_PASS_DICTIONARY_ALL = 30
    FTP_SAME_USER_PASS_DICTIONARY_ALL = 31
    CASSANDRA_SAME_USER_PASS_DICTIONARY_ALL = 32
    IRC_SAME_USER_PASS_DICTIONARY_ALL = 33
    MONGO_SAME_USER_PASS_DICTIONARY_ALL = 34
    MYSQL_SAME_USER_PASS_DICTIONARY_ALL = 35
    SMTP_SAME_USER_PASS_DICTIONARY_ALL = 36
    POSTGRES_SAME_USER_PASS_DICTIONARY_ALL = 37
    NETWORK_SERVICE_LOGIN = 38
    FIND_FLAG = 39
    NIKTO_WEB_HOST_SCAN = 40
    MASSCAN_HOST_SCAN = 41
    MASSCAN_ALL_SCAN = 42
    FIREWALK_HOST = 43
    FIREWALK_ALL = 44
    HTTP_ENUM_HOST = 45
    HTTP_ENUM_ALL = 46
    HTTP_GREP_HOST = 47
    HTTP_GREP_ALL = 48
    FINGER_HOST = 49
    FINGER_ALL = 50
    INSTALL_TOOLS = 51
    SSH_BACKDOOR = 52
    SAMBACRY_EXPLOIT = 53
    SHELLSHOCK_EXPLOIT = 54
    DVWA_SQL_INJECTION = 55
    CVE_2015_3306_EXPLOIT = 56
    CVE_2015_1427_EXPLOIT = 57
    CVE_2016_10033_EXPLOIT = 58
    CVE_2010_0426_PRIV_ESC = 59
    CVE_2015_5602_PRIV_ESC = 60
    STOP = 61
    CONTINUE = 62



from typing import Optional, List, Dict, Any
from csle_common.dao.emulation_observation.common.emulation_vulnerability_observation_state import \
    EmulationVulnerabilityObservationState
from csle_common.dao.emulation_config.transport_protocol import TransportProtocol
from csle_common.dao.emulation_config.credential import Credential
from csle_base.json_serializable import JSONSerializable


class NmapVuln(JSONSerializable):
    """
    DTO representing a vulnerability found with NMAP
    """

    def __init__(self, name: str, port: int, protocol: TransportProtocol, cvss: float, service: str,
                 credentials: Optional[List[Credential]] = None):
        """
        Initializes the DTO

        :param name: the name of the vulnerability
        :param port: the port of the vulnerability
        :param protocol: the protocol of the vulnerability
        :param cvss: the cvss of the vulnerability
        :param service: the service of the vulnerability
        :param credentials: the credentials of the vulnerability
        """
        self.name = name
        self.port = port
        self.protocol = protocol
        self.cvss = cvss
        self.service = service
        self.credentials = credentials
        if self.credentials is None:
            self.credentials = []

    def to_obs(self) -> EmulationVulnerabilityObservationState:
        """
        Converts the object into a VulnerabilityObservationState

        :return: the created VulnerabilityObservationState
        """
        service: Optional[str] = ""
        if self.credentials is None:
            raise ValueError("self.credentials is None")
        if len(self.credentials) > 0:
            service = self.credentials[0].service
        vuln = EmulationVulnerabilityObservationState(name=self.name, port=self.port, protocol=self.protocol,
                                                      cvss=self.cvss, credentials=self.credentials, service=service)
        return vuln

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        if self.credentials is None:
            raise ValueError("Credential list is None")
        return f"name:{self.name}, port:{self.port}, protocol:{self.protocol}, cvss:{self.cvss}, " \
               f"service:{self.service}, credentials:{list(map(lambda x: str(x), self.credentials))}"

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation

        :return: a dict representation of the object
        """
        d: Dict[str, Any] = {}
        d["name"] = self.name
        d["port"] = self.port
        d["protocol"] = self.protocol
        d["cvss"] = self.cvss
        d["service"] = self.service
        d["credentials"] = list(map(lambda x: x.to_dict(), self.credentials)) if \
            self.credentials is not None else self.credentials
        return d

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "NmapVuln":
        """
        Converts a dict representation to an instance

        :param d: the dict to convert
        :return: the created instance
        """
        obj = NmapVuln(
            name=d["name"], port=d["port"], protocol=d["protocol"], cvss=d["cvss"], service=d["service"],
            credentials=list(map(lambda x: Credential.from_dict(x), d["credentials"]))
        )
        return obj

    @staticmethod
    def from_json_file(json_file_path: str) -> "NmapVuln":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return NmapVuln.from_dict(json.loads(json_str))

    def __hash__(self) -> int:
        """
        :return: a hash representation of the object
        """
        return hash(self.name) + 31 * hash(self.port) + 31 * hash(self.protocol.value) + 28 * hash(self.service)



from typing import Dict, Any, List, Optional
from docker.models.containers import Container
from csle_base.json_serializable import JSONSerializable


class DockerContainerMetadata(JSONSerializable):
    """
    DTO Object representing a running or stopped Docker container
    """

    def __init__(self, name: str, status: str, short_id: str, image_short_id: str, image_tags: List[str], id: str,
                 created: str, ip: str, network_id: int, gateway: str, mac: str, ip_prefix_len: int,
                 name2: str, level: str, hostname: str, image_name: str, net: str,
                 dir: str, config_path: str, container_handle: Optional[Container],
                 emulation: str, kafka_container: str) -> None:
        """
        Initializes the DTO

        :param name: the name of the container
        :param status: the status of the container
        :param short_id: the short id of the container
        :param image_short_id: the short id of the container's image
        :param image_tags: the tags of the container's image
        :param id: the id of the container
        :param created: the time the container was created
        :param ip: the ip of the container
        :param network_id: the network id of the container
        :param gateway: the gateway of the container
        :param mac: the mac address of the container
        :param ip_prefix_len: the ip prefix len of the container
        :param name2: the name2 of the container
        :param level: the level of the container
        :param hostname: the hostname of the container
        :param image_name: the image name of the container
        :param net: the network of the container
        :param dir: the directory of the container
        :param config_path: the container configuration ṕath of the container
        :param container_handle: the py-docker container handle
        :param emulation: the emulation name
        :param kafka_container: the kafka container name
        """
        self.name = name
        self.status = status
        self.short_id = short_id
        self.image_short_id = image_short_id
        self.image_tags = image_tags
        self.id = id
        self.created = created
        self.ip = ip
        self.network_id = network_id
        self.gateway = gateway
        self.mac = mac
        self.ip_prefix_len = ip_prefix_len
        self.name2 = name2
        self.level = level
        self.hostname = hostname
        self.image_name = image_name
        self.net = net
        self.dir = dir
        self.config_path = config_path
        self.container_handle = container_handle
        self.emulation = emulation
        self.kafka_container = kafka_container

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "DockerContainerMetadata":
        """
        Converts a dict representation to an instance

        :param d: the dict to convert
        :return: the converted instance
        """
        obj = DockerContainerMetadata(
            name=d["name"],
            status=d["status"],
            short_id=d["short_id"],
            image_short_id=d["image_short_id"],
            image_tags=d["image_tags"],
            id=d["id"],
            created=d["created"],
            ip=d["ip"],
            network_id=d["network_id"],
            gateway=d["gateway"],
            mac=d["mac"],
            ip_prefix_len=d["ip_prefix_len"],
            name2=d["name2"],
            level=d["level"],
            hostname=d["hostname"],
            image_name=d["image_name"],
            net=d["net"],
            dir=d["dir"],
            config_path=d["config_path"],
            container_handle=d["container_handle"],
            emulation=d["emulation"],
            kafka_container=d["kafka_container"]
        )
        return obj

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation
        
        :return: a dict representation of the object
        """
        d: Dict[str, Any] = {}
        d["name"] = self.name
        d["status"] = self.status
        d["short_id"] = self.short_id
        d["image_short_id"] = self.image_short_id
        d["image_tags"] = self.image_tags
        d["id"] = self.id
        d["created"] = self.created
        d["ip"] = self.ip
        d["network_id"] = self.network_id
        d["gateway"] = self.gateway
        d["mac"] = self.mac
        d["ip_prefix_len"] = self.ip_prefix_len
        d["name2"] = self.name2
        d["level"] = self.level
        d["hostname"] = self.hostname
        d["image_name"] = self.image_name
        d["net"] = self.net
        d["config_path"] = self.config_path
        d["emulation"] = self.emulation
        d["kafka_container"] = self.kafka_container
        d["dir"] = self.dir
        d["container_handle"] = None
        return d

    def __str__(self):
        """
        :return: a string representation of the object
        """
        return f"name: {self.name}, status: {self.status}, short_id: {self.short_id}, " \
               f"image_short_id: {self.image_short_id}, image_tags: {self.image_tags}, id: {self.id}, " \
               f"created: {self.created}, ip: {self.ip}, network_id: {self.network_id}, gateway: {self.gateway}," \
               f"mac: {self.mac}, ip_prefix_len: {self.ip_prefix_len}, name2: {self.name2}," \
               f"level: {self.level}, hostname: {self.hostname}, image_name: {self.image_name}, net: {self.net}, " \
               f"dir: {self.dir}, config_path: {self.config_path}, emulation: {self.emulation}, " \
               f"kafka_container:{self.kafka_container}"

    @staticmethod
    def from_json_file(json_file_path: str) -> "DockerContainerMetadata":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return DockerContainerMetadata.from_dict(json.loads(json_str))



from csle_common.util.general_util import GeneralUtil


class TestShellUtilSuite:
    """
    Test suite for general_util.py
    """

    def test_execute_service_login_helper(self) -> None:
        """
        Tests the replace_first_octet_of_ip function

        :return: None
        """
        assert GeneralUtil.replace_first_octet_of_ip(ip="127.0.0.1", ip_first_octet=1) == "1.0.0.1"
        assert GeneralUtil.replace_first_octet_of_ip(ip="<execution_id>.0.0.1", ip_first_octet=19) == "19.0.0.1"
        assert GeneralUtil.replace_first_octet_of_ip(ip="<execution_id>.125.251.19", ip_first_octet=128) == \
               "128.125.251.19"



from typing import List, Set, Dict, Any, Optional
import copy
from csle_common.dao.emulation_config.flag import Flag
from csle_common.dao.emulation_config.credential import Credential
from csle_common.dao.emulation_observation.common.emulation_port_observation_state import EmulationPortObservationState
from csle_common.dao.emulation_observation.common.emulation_vulnerability_observation_state \
    import EmulationVulnerabilityObservationState
from csle_common.dao.emulation_observation.common.emulation_connection_observation_state \
    import EmulationConnectionObservationState
from csle_common.dao.emulation_action_result.nmap_host_result import NmapHostResult
from csle_common.dao.emulation_action_result.nmap_trace import NmapTrace
from csle_base.json_serializable import JSONSerializable


class EmulationAttackerMachineObservationState(JSONSerializable):
    """
    Represents the attacker's belief state of a component in the emulation
    """

    def __init__(self, ips: List[str]):
        """
        Initializes the state

        :param ips: the ip of the machine
        """
        assert ips is not None and len(ips) > 0
        self.ips = ips
        self.os = "unknown"
        self.ports: List[EmulationPortObservationState] = []
        self.cve_vulns: List[EmulationVulnerabilityObservationState] = []
        self.osvdb_vulns: List[EmulationVulnerabilityObservationState] = []
        self.shell_access = False
        self.shell_access_credentials: List[Credential] = []
        self.backdoor_credentials: List[Credential] = []
        self.logged_in = False
        self.root = False
        self.flags_found: Set[Flag] = set()
        self.filesystem_searched = False
        self.untried_credentials = False
        self.ssh_connections: List[EmulationConnectionObservationState] = []
        self.ftp_connections: List[EmulationConnectionObservationState] = []
        self.telnet_connections: List[EmulationConnectionObservationState] = []
        self.logged_in_services: List[str] = []
        self.root_services: List[str] = []
        self.hostnames: List[str] = []
        self.trace: Optional[NmapTrace] = None
        self.telnet_brute_tried = False
        self.ssh_brute_tried = False
        self.ftp_brute_tried = False
        self.cassandra_brute_tried = False
        self.irc_brute_tried = False
        self.mongo_brute_tried = False
        self.mysql_brute_tried = False
        self.smtp_brute_tried = False
        self.postgres_brute_tried = False
        self.tools_installed = False
        self.backdoor_installed = False
        self.backdoor_tried = False
        self.install_tools_tried = False
        self.sambacry_tried = False
        self.shellshock_tried = False
        self.dvwa_sql_injection_tried = False
        self.cve_2015_3306_tried = False
        self.cve_2015_1427_tried = False
        self.cve_2016_10033_tried = False
        self.cve_2010_0426_tried = False
        self.cve_2015_5602_tried = False
        self.reachable: Set[str] = set()

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        return f"ips:{self.ips}, os:{self.os}, shell_access:{self.shell_access}, logged_in:{self.logged_in}, " \
               f"root:{self.root}, num_ports:{len(self.ports)}, num_cve_vuln:{len(self.cve_vulns)}, " \
               f"num_cred:{len(self.shell_access_credentials)}," \
               f"num_ssh_connections:{len(self.ssh_connections)}," \
               f"num_ftp_connections:{len(self.ftp_connections)}, " \
               f"num_telnet_connections:{len(self.telnet_connections)}, " \
               f"num_osvdb_vuln:{len(self.osvdb_vulns)}, hostnames:{self.hostnames}, trace:{self.trace}, " \
               f"filesystem_searched:{self.filesystem_searched}, " \
               f"telnet_brute_tried:{self.telnet_brute_tried}, ssh_brute_tried:{self.ssh_brute_tried}, " \
               f"ftp_brute_tried:{self.ftp_brute_tried}," \
               f"cassandra_brute_tried:{self.cassandra_brute_tried}, irc_brute_tried:{self.irc_brute_tried}, " \
               f"mongo_brute_tried:{self.mongo_brute_tried}, mysql_brute_tried:{self.mysql_brute_tried}," \
               f"smtp_brute_tried:{self.smtp_brute_tried}, postgres_brute_tried:{self.postgres_brute_tried}, " \
               f"tools_installed:{self.tools_installed}, backdoor_installed:{self.backdoor_installed}," \
               f"num_backdoor_credentials:{len(self.backdoor_credentials)}, " \
               f"num_reachable_nodes:{len(self.reachable)}, " \
               f"backdoor_tried:{self.backdoor_tried}, install_tools_tried:{self.install_tools_tried}, " \
               f"sambacry_tried:{self.sambacry_tried}, shellshock_tried:{self.shellshock_tried}, " \
               f"dvwa_sql_injection_tried:{self.dvwa_sql_injection_tried}, " \
               f"cve_2015_3306_tried:{self.cve_2015_3306_tried}," \
               f"cve_2015_1427_tried:{self.cve_2015_1427_tried}, cve_2016_10033_tried:{self.cve_2016_10033_tried}, " \
               f"cve_2010_0426_tried:{self.cve_2010_0426_tried}, cve_2015_5602_tried:{self.cve_2015_5602_tried}," \
               f"flags_found:{self.flags_found}"

    def sort_ports(self) -> None:
        """
        Sorts the list of ports
        :return: None
        """
        for p in self.ports:
            p.port = int(p.port)
        self.ports = sorted(self.ports, key=lambda x: x.kafka_port, reverse=False)

    def sort_cve_vuln(self, vuln_lookup) -> None:
        """
        Sorts the list of vulnerabilities

        :param vuln_lookup: a lookup table for converting vulnerabilities between ids and names
        :return: None
        """
        self.cve_vulns = sorted(self.cve_vulns, key=lambda x: self._vuln_lookup(name=x.name, lookup_table=vuln_lookup),
                                reverse=False)

    def sort_shell_access(self, service_lookup) -> None:
        """
        Sorts the list of shell access credentials

        :param service_lookup: a lookup table for converting between service names and service ids
        :return: None
        """
        self.shell_access_credentials = sorted(
            self.shell_access_credentials,
            key=lambda x: service_lookup[x.service] if x.service is not None else x.username,
            reverse=False)

    def _vuln_lookup(self, name: str, lookup_table: Dict[str, int]) -> int:
        """
        Looks up the id of a vulnerability in a lookup table

        :param name: the name of the vulnerability
        :param lookup_table: the lookup table
        :return: the id of the vulnerability
        """
        if name in lookup_table:
            return lookup_table[name]
        else:
            return lookup_table["unknown"]

    def sort_osvdb_vuln(self) -> None:
        """
        Sorts the OSVDB vulnerabilities

        :return: None
        """
        self.osvdb_vulns = sorted(self.osvdb_vulns, key=lambda x: x.osvdb_id, reverse=False)

    def cleanup(self):
        """
        Cleans up environment state. This method is particularly useful in emulation mode where there are
        SSH/Telnet/FTP... connections that should be cleaned up, as well as background threads.

        :return: None
        """
        for c in self.ssh_connections:
            c.cleanup()
        for c in self.ftp_connections:
            c.cleanup()
        for c in self.telnet_connections:
            c.cleanup()

    def copy(self) -> "EmulationAttackerMachineObservationState":
        """
        :return: a copy of the DTO
        """
        m_copy = EmulationAttackerMachineObservationState(ips=self.ips)
        m_copy.os = self.os
        m_copy.ports = copy.deepcopy(self.ports)
        m_copy.cve_vulns = copy.deepcopy(self.cve_vulns)
        m_copy.osvdb_vulns = copy.deepcopy(self.osvdb_vulns)
        m_copy.shell_access = self.shell_access
        m_copy.shell_access_credentials = copy.deepcopy(self.shell_access_credentials)
        m_copy.backdoor_credentials = copy.deepcopy(self.backdoor_credentials)
        m_copy.logged_in = self.logged_in
        m_copy.root = self.root
        m_copy.flags_found = copy.deepcopy(self.flags_found)
        m_copy.filesystem_searched = self.filesystem_searched
        m_copy.untried_credentials = self.untried_credentials
        m_copy.ssh_connections = self.ssh_connections
        m_copy.ftp_connections = self.ftp_connections
        m_copy.telnet_connections = self.telnet_connections
        m_copy.logged_in_services = self.logged_in_services
        m_copy.root_services = self.root_services
        m_copy.hostnames = self.hostnames
        m_copy.trace = self.trace
        m_copy.telnet_brute_tried = self.telnet_brute_tried
        m_copy.ssh_brute_tried = self.ssh_brute_tried
        m_copy.ftp_brute_tried = self.ftp_brute_tried
        m_copy.cassandra_brute_tried = self.cassandra_brute_tried
        m_copy.irc_brute_tried = self.irc_brute_tried
        m_copy.mongo_brute_tried = self.mongo_brute_tried
        m_copy.mysql_brute_tried = self.mysql_brute_tried
        m_copy.smtp_brute_tried = self.smtp_brute_tried
        m_copy.postgres_brute_tried = self.postgres_brute_tried
        m_copy.tools_installed = self.tools_installed
        m_copy.backdoor_installed = self.backdoor_installed
        m_copy.backdoor_tried = self.backdoor_tried
        m_copy.install_tools_tried = self.install_tools_tried
        m_copy.reachable = self.reachable
        m_copy.sambacry_tried = self.sambacry_tried
        m_copy.shellshock_tried = self.shellshock_tried
        m_copy.dvwa_sql_injection_tried = self.dvwa_sql_injection_tried
        m_copy.cve_2015_3306_tried = self.cve_2015_3306_tried
        m_copy.cve_2015_1427_tried = self.cve_2015_1427_tried
        m_copy.cve_2016_10033_tried = self.cve_2016_10033_tried
        m_copy.cve_2010_0426_tried = self.cve_2010_0426_tried
        m_copy.cve_2015_5602_tried = self.cve_2015_5602_tried
        return m_copy

    @staticmethod
    def from_nmap_result(nmap_host_result: NmapHostResult) -> "EmulationAttackerMachineObservationState":
        """
        Converts the NmapHostResultDTO into a AttackerMachineObservationState

        :return: the created AttackerMachineObservationState
        """
        if nmap_host_result.ips is None or \
                nmap_host_result.ports is None or \
                nmap_host_result.vulnerabilities is None or \
                nmap_host_result.credentials is None or \
                nmap_host_result.hostnames is None:
            raise ValueError("NMapHostResult is incomplete, at least one attribute is None")
        m_obs = EmulationAttackerMachineObservationState(ips=nmap_host_result.ips)
        ports = list(map(lambda x: x.to_obs(), nmap_host_result.ports))
        m_obs.ports = ports
        if nmap_host_result.os is not None:
            m_obs.os = nmap_host_result.os.vendor.lower()
        vulnerabilities = list(map(lambda x: x.to_obs(), nmap_host_result.vulnerabilities))
        m_obs.cve_vulns = vulnerabilities
        credentials = list(map(lambda x: x.to_obs(), nmap_host_result.credentials))
        m_obs.shell_access_credentials = credentials
        if len(credentials) > 0:
            m_obs.shell_access = True
            m_obs.untried_credentials = True
        m_obs.hostnames = nmap_host_result.hostnames
        m_obs.trace = nmap_host_result.trace
        return m_obs

    def ips_match(self, ips: List[str]) -> bool:
        """
        Checks if a list of ips overlap with the ips of this machine

        :param ips: the list of ips to check
        :return:  True if they match, False otherwise
        """
        for ip in self.ips:
            if ip in ips:
                return True
            if ip.split(".")[-1] in list(map(lambda x: x.split(".")[-1], ips)):
                return True
        return False

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "EmulationAttackerMachineObservationState":
        """
        Converts a dict representation to an instance

        :param d: the dict to convert
        :return: the created instance
        """
        obj = EmulationAttackerMachineObservationState(ips=d["ips"])
        obj.os = d["os"]
        obj.ports = list(map(lambda x: EmulationPortObservationState.from_dict(x), d["ports"]))
        obj.cve_vulns = list(map(lambda x: EmulationVulnerabilityObservationState.from_dict(x), d["cve_vulns"]))
        obj.osvdb_vulns = list(map(lambda x: EmulationVulnerabilityObservationState.from_dict(x), d["osvdb_vulns"]))
        obj.shell_access = d["shell_access"]
        obj.shell_access_credentials = list(map(lambda x: Credential.from_dict(x), d["shell_access_credentials"]))
        obj.backdoor_credentials = list(map(lambda x: Credential.from_dict(x), d["backdoor_credentials"]))
        obj.logged_in = d["logged_in"]
        obj.root = d["root"]
        obj.flags_found = set(d["flags_found"])
        obj.filesystem_searched = d["filesystem_searched"]
        obj.untried_credentials = d["untried_credentials"]
        obj.logged_in_services = d["logged_in_services"]
        obj.root_services = d["root_services"]
        obj.hostnames = d["hostnames"]
        obj.trace = d["trace"]
        obj.telnet_brute_tried = d["telnet_brute_tried"]
        obj.ssh_brute_tried = d["ssh_brute_tried"]
        obj.ftp_brute_tried = d["ftp_brute_tried"]
        obj.cassandra_brute_tried = d["cassandra_brute_tried"]
        obj.irc_brute_tried = d["irc_brute_tried"]
        obj.mongo_brute_tried = d["mongo_brute_tried"]
        obj.mysql_brute_tried = d["mysql_brute_tried"]
        obj.smtp_brute_tried = d["smtp_brute_tried"]
        obj.postgres_brute_tried = d["postgres_brute_tried"]
        obj.tools_installed = d["tools_installed"]
        obj.backdoor_installed = d["backdoor_installed"]
        obj.backdoor_tried = d["backdoor_tried"]
        obj.install_tools_tried = d["install_tools_tried"]
        obj.sambacry_tried = d["sambacry_tried"]
        obj.shellshock_tried = d["shellshock_tried"]
        obj.dvwa_sql_injection_tried = d["dvwa_sql_injection_tried"]
        obj.cve_2015_3306_tried = d["cve_2015_3306_tried"]
        obj.cve_2015_1427_tried = d["cve_2015_1427_tried"]
        obj.cve_2016_10033_tried = d["cve_2016_10033_tried"]
        obj.cve_2010_0426_tried = d["cve_2010_0426_tried"]
        obj.cve_2015_5602_tried = d["cve_2015_5602_tried"]
        obj.reachable = set(d["reachable"])
        return obj

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation

        :return: a dict representation of the object
        """
        d: Dict[str, Any] = {}
        d["ips"] = self.ips
        d["os"] = self.os
        d["ports"] = list(map(lambda x: x.to_dict(), self.ports))
        d["cve_vulns"] = list(map(lambda x: x.to_dict(), self.cve_vulns))
        d["osvdb_vulns"] = list(map(lambda x: x.to_dict(), self.osvdb_vulns))
        d["shell_access"] = self.shell_access
        d["shell_access_credentials"] = list(map(lambda x: x.to_dict(), self.shell_access_credentials))
        d["backdoor_credentials"] = list(map(lambda x: x.to_dict(), self.backdoor_credentials))
        d["logged_in"] = self.logged_in
        d["root"] = self.root
        d["flags_found"] = list(self.flags_found)
        d["filesystem_searched"] = self.filesystem_searched
        d["untried_credentials"] = self.untried_credentials
        d["logged_in_services"] = self.logged_in_services
        d["root_services"] = self.root_services
        d["hostnames"] = self.hostnames
        if self.trace is None:
            d["trace"] = None
        else:
            d["trace"] = self.trace.to_dict()
        d["telnet_brute_tried"] = self.telnet_brute_tried
        d["ssh_brute_tried"] = self.ssh_brute_tried
        d["ftp_brute_tried"] = self.ftp_brute_tried
        d["cassandra_brute_tried"] = self.cassandra_brute_tried
        d["irc_brute_tried"] = self.irc_brute_tried
        d["mongo_brute_tried"] = self.mongo_brute_tried
        d["mysql_brute_tried"] = self.mysql_brute_tried
        d["smtp_brute_tried"] = self.smtp_brute_tried
        d["postgres_brute_tried"] = self.postgres_brute_tried
        d["tools_installed"] = self.tools_installed
        d["backdoor_installed"] = self.backdoor_installed
        d["backdoor_tried"] = self.backdoor_tried
        d["install_tools_tried"] = self.install_tools_tried
        d["sambacry_tried"] = self.sambacry_tried
        d["shellshock_tried"] = self.shellshock_tried
        d["dvwa_sql_injection_tried"] = self.dvwa_sql_injection_tried
        d["cve_2015_3306_tried"] = self.cve_2015_3306_tried
        d["cve_2015_1427_tried"] = self.cve_2015_1427_tried
        d["cve_2016_10033_tried"] = self.cve_2016_10033_tried
        d["cve_2010_0426_tried"] = self.cve_2010_0426_tried
        d["cve_2015_5602_tried"] = self.cve_2015_5602_tried
        d["reachable"] = list(self.reachable)
        return d

    @staticmethod
    def from_json_file(json_file_path: str) -> "EmulationAttackerMachineObservationState":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return EmulationAttackerMachineObservationState.from_dict(json.loads(json_str))

    def num_attributes(self) -> int:
        """
        :return: number of attributes of the DTO
        """
        attributes_per_emulation_port_obs_state = 0
        attributes_per_emulation_vuln_obs_state = 0
        attributes_per_credential = 0
        attributes_per_connection_obs_state = 0
        if len(self.ports) > 0:
            attributes_per_emulation_port_obs_state = self.ports[0].num_attributes()
        if len(self.cve_vulns) > 0:
            attributes_per_emulation_vuln_obs_state = self.cve_vulns[0].num_attributes()
        if len(self.cve_vulns) == 0 and len(self.osvdb_vulns) > 0:
            attributes_per_emulation_vuln_obs_state = self.osvdb_vulns[0].num_attributes()
        if len(self.shell_access_credentials) > 0:
            attributes_per_credential = self.shell_access_credentials[0].num_attributes()
        if len(self.ssh_connections) > 0:
            attributes_per_connection_obs_state = self.ssh_connections[0].num_attributes()
        if len(self.ssh_connections) == 0 and len(self.ftp_connections) > 0:
            attributes_per_connection_obs_state = self.ftp_connections[0].num_attributes()
        if len(self.ssh_connections) == 0 and len(self.ftp_connections) == 0 and len(self.telnet_connections) > 0:
            attributes_per_connection_obs_state = self.telnet_connections[0].num_attributes()

        return (len(self.ports) * attributes_per_emulation_port_obs_state +
                len(self.cve_vulns) * attributes_per_emulation_vuln_obs_state +
                len(self.osvdb_vulns) * attributes_per_emulation_vuln_obs_state +
                len(self.shell_access_credentials) * attributes_per_credential +
                len(self.backdoor_credentials) * attributes_per_credential +
                len(self.ssh_connections) * attributes_per_connection_obs_state +
                len(self.ftp_connections) * attributes_per_connection_obs_state +
                len(self.telnet_connections) * attributes_per_connection_obs_state +
                len(self.logged_in_services) + len(self.hostnames) + len(self.root_services) +
                len(self.hostnames) + len(self.reachable) + len(self.flags_found) + 29)

    @staticmethod
    def schema() -> "EmulationAttackerMachineObservationState":
        """
        :return: get the schema of the DTO
        """
        dto = EmulationAttackerMachineObservationState(ips=[""])
        dto.ports = [EmulationPortObservationState.schema()]
        dto.cve_vulns = [EmulationVulnerabilityObservationState.schema()]
        dto.osvdb_vulns = [EmulationVulnerabilityObservationState.schema()]
        dto.shell_access_credentials = [Credential.schema()]
        dto.backdoor_credentials = [Credential.schema()]
        dto.flags_found.add(Flag.schema())
        dto.ssh_connections = [EmulationConnectionObservationState.schema()]
        dto.ftp_connections = [EmulationConnectionObservationState.schema()]
        dto.telnet_connections = [EmulationConnectionObservationState.schema()]
        dto.logged_in_services = [""]
        dto.root_services = [""]
        dto.hostnames = [""]
        dto.trace = NmapTrace.schema()
        dto.reachable.add("")
        return dto



import logging
import subprocess
import time
import csle_common.constants.constants as constants
from csle_common.dao.emulation_config.containers_config import ContainersConfig
from csle_common.dao.emulation_config.emulation_env_config import EmulationEnvConfig
from csle_common.util.emulation_util import EmulationUtil


class OVSController:
    """
    Class that contains functionality for interacting and managing OVS bridges in CSLE
    """

    @staticmethod
    def create_virtual_switches_on_container(containers_config: ContainersConfig, physical_server_ip: str,
                                             logger: logging.Logger) -> None:
        """
        Sets up OVS switches on containers

        :param containers_config: the containers configuration
        :param physical_server_ip: the ip of the physical server
        :param logger: the logger to use for logging
        :return: None
        """
        for c in containers_config.containers:
            if c.physical_host_ip != physical_server_ip:
                continue
            for ovs_image in constants.CONTAINER_IMAGES.OVS_IMAGES:
                if ovs_image in c.name:
                    logger.info(f"Creating OVS bridge and ports "
                                f"on container: {c.get_full_name()} with IP {c.docker_gw_bridge_ip} ({c.get_ips()[0]})")
                    container_name = c.get_full_name()
                    bridge_name = constants.OVS.DEFAULT_BRIDGE_NAME
                    cmd = f"{constants.COMMANDS.DOCKER_EXEC_COMMAND} {container_name} {constants.OVS.OVS_VSCTL} " \
                          f"{constants.OVS.ADD_BR} {bridge_name}"
                    logger.info(f"Running cmd: {cmd} on container: {c.get_full_name()} "
                                f"with IP {c.docker_gw_bridge_ip} ({c.get_ips()[0]})")
                    subprocess.Popen(cmd, stdout=subprocess.DEVNULL, shell=True)
                    time.sleep(1)
                    cmd = f"{constants.COMMANDS.DOCKER_EXEC_COMMAND} {container_name} ifconfig {bridge_name} up"
                    logger.info(f"Running cmd: {cmd} on container: {c.get_full_name()} "
                                f"with IP {c.docker_gw_bridge_ip} ({c.get_ips()[0]})")
                    subprocess.Popen(cmd, stdout=subprocess.DEVNULL, shell=True)
                    time.sleep(1)
                    idx = 0
                    for ip_net in c.ips_and_networks:
                        ip, net = ip_net
                        cmd = f"{constants.COMMANDS.DOCKER_EXEC_COMMAND} {container_name} {constants.OVS.OVS_VSCTL} " \
                              f"{constants.OVS.ADD_PORT} {bridge_name} {net.interface}"
                        logger.info(f"Running cmd: {cmd} on container: {c.get_full_name()} "
                                    f"with IP {c.docker_gw_bridge_ip} ({c.get_ips()[0]})")
                        subprocess.Popen(cmd, stdout=subprocess.DEVNULL, shell=True)
                        time.sleep(1)
                        cmd = f"{constants.COMMANDS.DOCKER_EXEC_COMMAND} {container_name} ifconfig {net.interface} 0"
                        logger.info(f"Running cmd: {cmd} on container: {c.get_full_name()} "
                                    f"with IP {c.docker_gw_bridge_ip} ({c.get_ips()[0]})")
                        subprocess.Popen(cmd, stdout=subprocess.DEVNULL, shell=True)
                        time.sleep(1)
                        if idx == 0:
                            bridge_intf = bridge_name
                        else:
                            bridge_intf = f"{bridge_name}:{idx}"
                        cmd = f"{constants.COMMANDS.DOCKER_EXEC_COMMAND} {container_name} " \
                              f"ifconfig {bridge_intf} {ip} netmask {net.bitmask}"
                        logger.info(f"Running cmd: {cmd} on container: {c.get_full_name()} "
                                    f"with IP {c.docker_gw_bridge_ip} ({c.get_ips()[0]})")
                        subprocess.Popen(cmd, stdout=subprocess.DEVNULL, shell=True)
                        time.sleep(1)
                        idx += 1

    @staticmethod
    def apply_ovs_config(emulation_env_config: EmulationEnvConfig, physical_server_ip: str,
                         logger: logging.Logger) -> None:
        """
        Aplies the OVS configuration on the OVS switches

        :param emulation_env_config: the emulation env config
        :param physical_server_ip: ip of the physical server
        :param logger: the logger to use for logging
        :return: None
        """
        for ovs_sw in emulation_env_config.ovs_config.switch_configs:
            if ovs_sw.physical_host_ip != physical_server_ip:
                logger.info(f"Wrong host IP of switch: {ovs_sw.container_name}, {ovs_sw.physical_host_ip}, "
                            f"{physical_server_ip}")
                continue
            logger.info(f"Configuring OVS bridge on container: {ovs_sw.container_name}, "
                        f"physical host: {ovs_sw.physical_host_ip}, gw ip: {ovs_sw.docker_gw_bridge_ip} "
                        f"container IP: {ovs_sw.ip}")
            EmulationUtil.connect_admin(emulation_env_config=emulation_env_config, ip=ovs_sw.docker_gw_bridge_ip)

            bridge_name = constants.OVS.DEFAULT_BRIDGE_NAME
            cmd = f"{constants.COMMANDS.SUDO} {constants.OVS.OVS_VSCTL} set bridge {bridge_name} " \
                  f"protocols={','.join(ovs_sw.openflow_protocols)}"
            logger.info(f"Running cmd:{cmd} on container: {ovs_sw.container_name}")
            EmulationUtil.execute_ssh_cmd(cmd=cmd, conn=emulation_env_config.connections[ovs_sw.docker_gw_bridge_ip])
            logger.info(f"Command: {cmd} completed")
            time.sleep(1)
            cmd = f"{constants.COMMANDS.SUDO} {constants.OVS.OVS_VSCTL} set-controller {bridge_name} " \
                  f"{ovs_sw.controller_transport_protocol}:{ovs_sw.controller_ip}:{ovs_sw.controller_port}"
            logger.info(f"Running cmd:{cmd} on container: {ovs_sw.container_name}")
            EmulationUtil.execute_ssh_cmd(cmd=cmd, conn=emulation_env_config.connections[ovs_sw.docker_gw_bridge_ip])
            logger.info(f"Command: {cmd} completed")



import logging
from typing import List
import grpc
import time
import csle_collector.ryu_manager.ryu_manager_pb2_grpc
import csle_collector.ryu_manager.ryu_manager_pb2
import csle_collector.ryu_manager.query_ryu_manager
import csle_collector.ryu_manager.ryu_manager_util
import csle_common.constants.constants as constants
import csle_collector.constants.constants as collector_constants
from csle_common.dao.emulation_config.emulation_env_config import EmulationEnvConfig
from csle_common.dao.emulation_config.ryu_managers_info import RyuManagersInfo
from csle_common.util.emulation_util import EmulationUtil


class SDNControllerManager:
    """
    Class managing interaction with the SDN controller
    """

    @staticmethod
    def start_ryu_manager(emulation_env_config: EmulationEnvConfig, logger: logging.Logger) -> None:
        """
        Utility method for starting the Ryu manager

        :param emulation_env_config: the emulation env config
        :param logger: the logger to use for logging
        :return: None
        """
        if emulation_env_config.sdn_controller_config is None:
            raise ValueError(f"Cannot start ryu manager for emulation without an SDN config. "
                             f"Emulation: {emulation_env_config.name}")
        # Check if ryu_manager is already running
        status = None
        not_running = False
        try:
            status = SDNControllerManager.get_ryu_status_by_port_and_ip(
                ip=emulation_env_config.sdn_controller_config.container.docker_gw_bridge_ip,
                port=emulation_env_config.sdn_controller_config.manager_port, timeout=5)
        except Exception:
            not_running = True
        status_str = ""
        if status is None:
            not_running = True
        else:
            status_str = (f"ryu_running: {status.ryu_running}, monitor_running: {status.monitor_running}, "
                          f"port: {status.port}, web_port: {status.web_port}, controller: {status.controller}, "
                          f"kafka_ip: {status.kafka_ip}, kafka_port: {status.kafka_port}, "
                          f"time_step_len: {status.time_step_len}")

        if not_running:
            logger.info(
                f"Starting ryu manager on node: "
                f"{emulation_env_config.sdn_controller_config.container.docker_gw_bridge_ip} "
                f"({emulation_env_config.sdn_controller_config.container.get_ips()}, "
                f"{emulation_env_config.sdn_controller_config.container.get_full_name()})")

            # Connect
            EmulationUtil.connect_admin(emulation_env_config=emulation_env_config,
                                        ip=emulation_env_config.sdn_controller_config.container.docker_gw_bridge_ip,
                                        create_producer=False)

            # Stop old background job if running
            cmd = (constants.COMMANDS.SUDO + constants.COMMANDS.SPACE_DELIM + constants.COMMANDS.PKILL +
                   constants.COMMANDS.SPACE_DELIM + constants.TRAFFIC_COMMANDS.RYU_MANAGER_FILE_NAME)
            o, e, _ = EmulationUtil.execute_ssh_cmd(
                cmd=cmd,
                conn=emulation_env_config.get_connection(
                    ip=emulation_env_config.sdn_controller_config.container.docker_gw_bridge_ip))

            # Start the ryu_manager
            cmd = constants.COMMANDS.START_RYU_MANAGER.format(
                emulation_env_config.sdn_controller_config.manager_port,
                emulation_env_config.sdn_controller_config.manager_log_dir,
                emulation_env_config.sdn_controller_config.manager_log_file,
                emulation_env_config.sdn_controller_config.manager_max_workers)
            o, e, _ = EmulationUtil.execute_ssh_cmd(
                cmd=cmd,
                conn=emulation_env_config.get_connection(
                    ip=emulation_env_config.sdn_controller_config.container.docker_gw_bridge_ip))
            logger.info(f"Starting the Ryu manager, cmd: {cmd}, out:{o.decode()}, err: {e.decode()}")
            time.sleep(2)
        else:
            logger.info(
                f"Ryu manager was already running on node: "
                f"{emulation_env_config.sdn_controller_config.container.docker_gw_bridge_ip} "
                f"({emulation_env_config.sdn_controller_config.container.get_ips()}, "
                f"{emulation_env_config.sdn_controller_config.container.get_full_name()}). "
                f"Status: {status_str}")

    @staticmethod
    def stop_ryu_manager(emulation_env_config: EmulationEnvConfig, logger: logging.Logger) -> None:
        """
        Utility method for stopping the Ryu manager

        :param emulation_env_config: the emulation env config
        :param logger: the logger to use for logging
        :return: None
        """
        if emulation_env_config.sdn_controller_config is None:
            raise ValueError(f"Cannot stop ryu manager for emulation without an SDN config. "
                             f"Emulation: {emulation_env_config.name}")
        # Connect
        EmulationUtil.connect_admin(emulation_env_config=emulation_env_config,
                                    ip=emulation_env_config.sdn_controller_config.container.docker_gw_bridge_ip,
                                    create_producer=False)

        logger.info(
            f"Stopping ryu manager on node: "
            f"{emulation_env_config.sdn_controller_config.container.docker_gw_bridge_ip} "
            f"({emulation_env_config.sdn_controller_config.container.get_ips()[0]}, "
            f"{emulation_env_config.sdn_controller_config.container.get_full_name()})")

        # Stop background job
        cmd = (constants.COMMANDS.SUDO + constants.COMMANDS.SPACE_DELIM + constants.COMMANDS.PKILL +
               constants.COMMANDS.SPACE_DELIM + constants.TRAFFIC_COMMANDS.RYU_MANAGER_FILE_NAME)
        o, e, _ = EmulationUtil.execute_ssh_cmd(
            cmd=cmd,
            conn=emulation_env_config.get_connection(
                ip=emulation_env_config.sdn_controller_config.container.docker_gw_bridge_ip))
        logger.info(f"Stopping the Ryu manager, cmd: {cmd}, out:{o.decode()}, err: {e.decode()}")
        time.sleep(2)

    @staticmethod
    def get_ryu_status(emulation_env_config: EmulationEnvConfig, logger: logging.Logger) -> \
            csle_collector.ryu_manager.ryu_manager_pb2.RyuDTO:
        """
        Method for querying the RyuManager about the status of the Ryu SDN controller

        :param emulation_env_config: the emulation config
        :param logger: the logger to use for logging
        :return: a RyuDTO with the status
        """
        if emulation_env_config.sdn_controller_config is None:
            raise ValueError(f"Cannot get ryu status for emulation without an SDN config. "
                             f"Emulation: {emulation_env_config.name}")
        SDNControllerManager.start_ryu_manager(emulation_env_config=emulation_env_config, logger=logger)
        ryu_dto = SDNControllerManager.get_ryu_status_by_port_and_ip(
            ip=emulation_env_config.sdn_controller_config.container.docker_gw_bridge_ip,
            port=emulation_env_config.sdn_controller_config.manager_port)
        return ryu_dto

    @staticmethod
    def get_ryu_status_by_port_and_ip(ip: str, port: int, timeout: int = collector_constants.GRPC.TIMEOUT_SECONDS) \
            -> csle_collector.ryu_manager.ryu_manager_pb2.RyuDTO:
        """
        Method for querying the RyuManager about the status of the Ryu SDN controller

        :param ip: the ip where the RyuManager is running
        :param port: the port the RyuManager is listening to
        :param timeout: the timeout of the GRPC query
        :return: an RyuDTO with the status
        """
        # Open a gRPC session
        with grpc.insecure_channel(f'{ip}:{port}', options=constants.GRPC_SERVERS.GRPC_OPTIONS) as channel:
            stub = csle_collector.ryu_manager.ryu_manager_pb2_grpc.RyuManagerStub(channel)
            ryu_dto = csle_collector.ryu_manager.query_ryu_manager.get_ryu_status(stub, timeout=timeout)
            return ryu_dto

    @staticmethod
    def stop_ryu(emulation_env_config: EmulationEnvConfig, logger: logging.Logger) \
            -> csle_collector.ryu_manager.ryu_manager_pb2.RyuDTO:
        """
        Method for requesting the RyuManager to stop the RYU SDN controller

        :param emulation_env_config: the emulation env config
        :return: a RyuDTO with the status of the server
        """
        if emulation_env_config.sdn_controller_config is None:
            raise ValueError(f"Cannot stop Ryu for emulation without an SDN config. "
                             f"Emulation: {emulation_env_config.name}")
        logger.info(
            f"Stopping RYU SDN controller on container: "
            f"{emulation_env_config.sdn_controller_config.container.docker_gw_bridge_ip} "
            f"({emulation_env_config.sdn_controller_config.container.get_ips()[0]}, "
            f"{emulation_env_config.sdn_controller_config.container.get_full_name()})")
        SDNControllerManager.start_ryu_manager(emulation_env_config=emulation_env_config, logger=logger)

        # Open a gRPC session
        with grpc.insecure_channel(
                f'{emulation_env_config.sdn_controller_config.container.docker_gw_bridge_ip}:'
                f'{emulation_env_config.sdn_controller_config.manager_port}',
                options=constants.GRPC_SERVERS.GRPC_OPTIONS) as channel:
            stub = csle_collector.ryu_manager.ryu_manager_pb2_grpc.RyuManagerStub(channel)
            ryu_dto = csle_collector.ryu_manager.query_ryu_manager.stop_ryu(stub)
            return ryu_dto

    @staticmethod
    def start_ryu(emulation_env_config: EmulationEnvConfig, physical_server_ip: str, logger: logging.Logger) \
            -> csle_collector.ryu_manager.ryu_manager_pb2.RyuDTO:
        """
        Method for requesting the RyuManager to start the Ryu SDN controller

        :param emulation_env_config: the emulation env config
        :param physical_server_ip: the ip of the physical server
        :param logger: the logger to use for logging
        :return: an RyuDTO with the status of the server
        """
        if emulation_env_config.sdn_controller_config is None:
            raise ValueError(f"Cannot start Ryu for emulation without an SDN config. "
                             f"Emulation: {emulation_env_config.name}")
        if emulation_env_config.sdn_controller_config.container.physical_host_ip != physical_server_ip:
            return
        logger.info(
            f"Starting Ryu SDN controller on container: "
            f"{emulation_env_config.sdn_controller_config.container.docker_gw_bridge_ip} "
            f"({emulation_env_config.sdn_controller_config.container.get_ips()[0]}, "
            f"{emulation_env_config.sdn_controller_config.container.get_full_name()})")
        SDNControllerManager.start_ryu_manager(emulation_env_config=emulation_env_config, logger=logger)

        # Open a gRPC session
        with grpc.insecure_channel(
                f'{emulation_env_config.sdn_controller_config.container.docker_gw_bridge_ip}:'
                f'{emulation_env_config.sdn_controller_config.manager_port}',
                options=constants.GRPC_SERVERS.GRPC_OPTIONS) as channel:
            stub = csle_collector.ryu_manager.ryu_manager_pb2_grpc.RyuManagerStub(channel)
            logger.info(
                f"Starting RYU, port: {emulation_env_config.sdn_controller_config.controller_port}, "
                f"web_port: {emulation_env_config.sdn_controller_config.controller_web_api_port}, "
                f"controller: {emulation_env_config.sdn_controller_config.controller_module_name}")
            ryu_dto = csle_collector.ryu_manager.query_ryu_manager.start_ryu(
                stub, port=emulation_env_config.sdn_controller_config.controller_port,
                web_port=emulation_env_config.sdn_controller_config.controller_web_api_port,
                controller=emulation_env_config.sdn_controller_config.controller_module_name)
            return ryu_dto

    @staticmethod
    def start_ryu_monitor(emulation_env_config: EmulationEnvConfig, physical_server_ip: str, logger: logging.Logger) \
            -> csle_collector.ryu_manager.ryu_manager_pb2.RyuDTO:
        """
        Method for requesting the RyuManager to start the Ryu monitor

        :param emulation_env_config: the emulation env config
        :param physical_server_ip: ip of the physical server
        :param logger: the logger to use for logging
        :return: an RyuDTO with the status
        """
        if emulation_env_config.sdn_controller_config is None:
            raise ValueError(f"Cannot start Ryu monitor for emulation without an SDN config. "
                             f"Emulation: {emulation_env_config.name}")
        if emulation_env_config.sdn_controller_config.container.physical_host_ip != physical_server_ip:
            return
        logger.info(
            f"Starting the ryu monitor on container: "
            f"{emulation_env_config.sdn_controller_config.container.docker_gw_bridge_ip} "
            f"({emulation_env_config.sdn_controller_config.container.get_ips()[0]}, "
            f"{emulation_env_config.sdn_controller_config.container.get_full_name()})")
        SDNControllerManager.start_ryu_manager(emulation_env_config=emulation_env_config, logger=logger)

        # Open a gRPC session
        with grpc.insecure_channel(
                f'{emulation_env_config.sdn_controller_config.container.docker_gw_bridge_ip}:'
                f'{emulation_env_config.sdn_controller_config.manager_port}',
                options=constants.GRPC_SERVERS.GRPC_OPTIONS) as channel:
            stub = csle_collector.ryu_manager.ryu_manager_pb2_grpc.RyuManagerStub(channel)
            ryu_dto = csle_collector.ryu_manager.query_ryu_manager.start_ryu_monitor(
                stub, kafka_ip=emulation_env_config.kafka_config.container.get_ips()[0],
                kafka_port=emulation_env_config.kafka_config.kafka_port,
                time_step_len=emulation_env_config.sdn_controller_config.time_step_len_seconds)
            return ryu_dto

    @staticmethod
    def stop_ryu_monitor(emulation_env_config: EmulationEnvConfig, logger: logging.Logger) -> \
            csle_collector.ryu_manager.ryu_manager_pb2.RyuDTO:
        """
        Method for requesting the RyuManager to stop the ryu monitor

        :param emulation_env_config: the emulation env config
        :param logger: the logger to use for logging
        :return: an RyuDTO with the status of the server
        """
        if emulation_env_config.sdn_controller_config is None:
            raise ValueError(f"Cannot stop Ryu monitor for emulation without an SDN config. "
                             f"Emulation: {emulation_env_config.name}")
        logger.info(
            f"Stopping Ryu monitor on container: "
            f"{emulation_env_config.sdn_controller_config.container.docker_gw_bridge_ip} "
            f"({emulation_env_config.sdn_controller_config.container.get_ips()[0]}, "
            f"{emulation_env_config.sdn_controller_config.container.get_full_name()})")
        SDNControllerManager.start_ryu_manager(emulation_env_config=emulation_env_config, logger=logger)

        # Open a gRPC session
        with grpc.insecure_channel(
                f'{emulation_env_config.sdn_controller_config.container.docker_gw_bridge_ip}:'
                f'{emulation_env_config.sdn_controller_config.manager_port}',
                options=constants.GRPC_SERVERS.GRPC_OPTIONS) as channel:
            stub = csle_collector.ryu_manager.ryu_manager_pb2_grpc.RyuManagerStub(channel)
            ryu_dto = csle_collector.ryu_manager.query_ryu_manager.stop_ryu_monitor(stub)
            return ryu_dto

    @staticmethod
    def get_ryu_managers_ips(emulation_env_config: EmulationEnvConfig) -> List[str]:
        """
        A method that extracts the IPS of the Ryu managers in a given emulation

        :param emulation_env_config: the emulation env config
        :return: the list of IP addresses
        """
        if emulation_env_config.sdn_controller_config is None:
            raise ValueError(f"Cannot get Ryu managers IPs for emulation without an SDN config. "
                             f"Emulation: {emulation_env_config.name}")
        return [emulation_env_config.sdn_controller_config.container.docker_gw_bridge_ip]

    @staticmethod
    def get_ryu_managers_ports(emulation_env_config: EmulationEnvConfig) -> List[int]:
        """
        A method that extracts the ports of the Ryu managers in a given emulation

        :param emulation_env_config: the emulation env config
        :return: the list of IP addresses
        """
        if emulation_env_config.sdn_controller_config is None:
            raise ValueError(f"Cannot get Ryu managers ports for emulation without an SDN config. "
                             f"Emulation: {emulation_env_config.name}")
        return [emulation_env_config.sdn_controller_config.manager_port]

    @staticmethod
    def get_ryu_managers_info(emulation_env_config: EmulationEnvConfig, active_ips: List[str],
                              logger: logging.Logger, physical_server_ip: str) -> RyuManagersInfo:
        """
        Extracts the information of the Ryu managers for a given emulation

        :param emulation_env_config: the configuration of the emulation
        :param active_ips: list of active IPs
        :param logger: the logger to use for logging
        :param physical_server_ip: the ip of the physical server
        :return: a DTO with the status of the Ryu managers
        """
        if emulation_env_config.sdn_controller_config is None:
            raise ValueError(f"Cannot get ryu managers info for emulation without an SDN config. "
                             f"Emulation: {emulation_env_config.name}")
        ryu_managers_ips = SDNControllerManager.get_ryu_managers_ips(emulation_env_config=emulation_env_config)
        ryu_managers_ports = SDNControllerManager.get_ryu_managers_ports(emulation_env_config=emulation_env_config)
        ryu_managers_statuses = []
        ryu_managers_running = []
        for ip in ryu_managers_ips:
            if ip not in active_ips or not EmulationUtil.physical_ip_match(
                    emulation_env_config=emulation_env_config, ip=ip, physical_host_ip=physical_server_ip):
                continue
            status = None
            try:
                status = SDNControllerManager.get_ryu_status_by_port_and_ip(
                    port=emulation_env_config.sdn_controller_config.manager_port, ip=ip)
                running = True
            except Exception as e:
                running = False
                logger.debug(f"Could not fetch Ryu manager status on IP:{ip}, error: {str(e)}, {repr(e)}")
            if status is not None:
                ryu_managers_statuses.append(status)
            else:
                ryu_managers_statuses.append(csle_collector.ryu_manager.ryu_manager_util.RyuManagerUtil.ryu_dto_empty())
            ryu_managers_running.append(running)
        execution_id = emulation_env_config.execution_id
        emulation_name = emulation_env_config.name
        ryu_manager_info_dto = RyuManagersInfo(
            ryu_managers_running=ryu_managers_running, ips=ryu_managers_ips, execution_id=execution_id,
            emulation_name=emulation_name, ryu_managers_statuses=ryu_managers_statuses, ports=ryu_managers_ports,
            physical_server_ip=emulation_env_config.sdn_controller_config.container.physical_host_ip)
        return ryu_manager_info_dto



from typing import Tuple, List, Union
import time
import paramiko
import psutil
import csle_collector.constants.constants as collector_constants
import csle_common.constants.constants as constants
from csle_common.dao.emulation_config.emulation_env_config import EmulationEnvConfig
from csle_common.dao.emulation_action.attacker.emulation_attacker_action import EmulationAttackerAction
from csle_common.dao.emulation_observation.common.emulation_connection_observation_state \
    import EmulationConnectionObservationState
from csle_common.dao.emulation_config.emulation_env_state import EmulationEnvState
from csle_common.dao.emulation_action.attacker.emulation_attacker_action_id import EmulationAttackerActionId
from csle_common.util.env_dynamics_util import EnvDynamicsUtil
from csle_common.dao.emulation_action.attacker.emulation_attacker_action_type import EmulationAttackerActionType
from csle_common.dao.emulation_action.attacker.emulation_attacker_action_outcome import EmulationAttackerActionOutcome
from csle_common.dao.emulation_config.credential import Credential
from csle_common.dao.emulation_config.transport_protocol import TransportProtocol
from csle_common.util.ssh_util import SSHUtil
from csle_common.logging.log import Logger


class EmulationUtil:
    """
    Class containing utility functions for the emulation-middleware
    """

    @staticmethod
    def execute_ssh_cmds(cmds: List[str], conn, wait_for_completion: bool = True) -> List[Tuple[bytes, bytes, float]]:
        """
        Executes a list of commands over an ssh connection to the emulation

        :param cmds: the list of commands
        :param conn: the ssh connection
        :param wait_for_completion: whether to wait for completion of the commands or not
        :return: List of tuples (outdata, errdata, total_time)
        """
        return SSHUtil.execute_ssh_cmds(cmds=cmds, conn=conn, wait_for_completion=wait_for_completion)

    @staticmethod
    def execute_ssh_cmd(cmd: str, conn, wait_for_completion: bool = True) -> Tuple[bytes, bytes, float]:
        """
        Executes an action on the emulation over a ssh connection,
        this is a synchronous operation that waits for the completion of the action before returning

        :param cmd: the command to execute
        :param conn: the ssh connection
        :param wait_for_completion: boolean flag whether to wait for completion or not
        :return: outdata, errdata, total_time
        """
        return SSHUtil.execute_ssh_cmd(cmd=cmd, conn=conn, wait_for_completion=wait_for_completion)

    @staticmethod
    def log_measured_action_time(total_time, action: EmulationAttackerAction,
                                 emulation_env_config: EmulationEnvConfig, create_producer: bool = False) -> None:
        """
        Logs the measured time of an action to Kafka

        :param total_time: the total time of executing the action
        :param action: the action
        :param emulation_env_config: the environment config
        :param create_producer: boolean flag whether to create a producer from the csle-admin
        :return: None
        """
        action.execution_time = total_time
        record = action.to_kafka_record()
        if emulation_env_config.producer is None and create_producer:
            emulation_env_config.create_producer()
        if emulation_env_config.producer is not None:
            emulation_env_config.producer.produce(collector_constants.KAFKA_CONFIG.ATTACKER_ACTIONS_TOPIC_NAME, record)
            emulation_env_config.producer.poll(0)
            emulation_env_config.producer.flush()

    @staticmethod
    def _check_if_ssh_server_is_running(conn, telnet: bool = False) -> bool:
        """
        Checks if an ssh server is running on the machine

        :param conn: the connection to use for the command
        :param telnet: whether the connection is a telnet connection
        :return: True if server is running, else false
        """
        cmd = "service ssh status"
        if not telnet:
            outdata, errdata, total_time = EmulationUtil.execute_ssh_cmd(cmd=cmd, conn=conn)
            return "is running" in outdata.decode() or "is running" in errdata.decode()
        else:
            cmd = cmd + "\n"
            conn.write(cmd.encode())
            response = conn.read_until(constants.TELNET.PROMPT, timeout=5)
            return "is running" in response.decode()

    @staticmethod
    def _list_all_users(c: EmulationConnectionObservationState, emulation_env_config: EmulationEnvConfig,
                        telnet: bool = False) -> List[str]:
        """
        List all users on a machine

        :param c: the connection to user for the command
        :param telnet: whether it is a telnet connection
        :param emulation_env_config: emulation env config
        :return: list of users
        """
        cmd = constants.SHELL.LIST_ALL_USERS
        users = []
        for i in range(constants.ENV_CONSTANTS.ATTACKER_RETRY_FIND_USERS):
            if not telnet:
                outdata, errdata, total_time = EmulationUtil.execute_ssh_cmd(cmd=cmd, conn=c.conn)
                outdata_1 = outdata.decode()
                users = outdata_1.split("\n")
                users = list(filter(lambda x: x != '', users))
            else:
                cmd = cmd + "\n"
                c.conn.write(cmd.encode())
                response = c.conn.read_until(constants.TELNET.PROMPT, timeout=5)
                response = response.decode()
                users = response.split("\n")
                users = list(map(lambda x: x.replace("\r", ""), users))
                users = list(filter(lambda x: x != '', users))
            if len(users) == 0:
                continue
            else:
                break
        if len(users) == 0:
            raise ValueError("users empty, ip:{}, telnet:{}, root:{}, username:{}".format(c.ip, telnet, c.root,
                                                                                          c.username))
        return users

    @staticmethod
    def is_connection_active(conn):
        """
        Checks if a given connection is active or not

        :param conn: the connection to check
        :return: True if active, otherwise False
        """
        if conn is not None and conn.get_transport() is not None:
            return conn.get_transport().is_active()
        else:
            return False

    @staticmethod
    def setup_custom_connection(user: str, pw: str, source_ip: str, port: int, target_ip: str, proxy_conn,
                                root: bool) \
            -> Union[EmulationConnectionObservationState, None]:
        """
        Utility function for setting up a custom SSH connection given credentials and a proxy connection
        :param user: the username of the new connection
        :param pw: the pw of the new connection
        :param source_ip: the ip of the proxy
        :param port: the port of the new connection
        :param target_ip: the ip to connect to
        :param proxy_conn: the proxy connection
        :param root: whether it is a root connection or not
        :return: the new connection
        """
        cr = Credential(username=user, pw=pw, root=root, protocol=TransportProtocol.TCP,
                        service=constants.SSH.SERVICE_NAME, port=constants.SSH.DEFAULT_PORT)
        agent_addr = (source_ip, port)
        target_addr = (target_ip, port)
        agent_transport = proxy_conn.conn.get_transport()
        try:
            relay_channel = agent_transport.open_channel(constants.SSH.DIRECT_CHANNEL, target_addr, agent_addr,
                                                         timeout=3)
            target_conn = paramiko.SSHClient()
            target_conn.set_missing_host_key_policy(paramiko.AutoAddPolicy())
            target_conn.connect(target_ip, username=user, password=pw, sock=relay_channel)
            transport = target_conn.get_transport()
            if transport is not None:
                transport.set_keepalive(5)
            else:
                raise ValueError(f"Connection failed, ip: {target_ip}, user: {user}, pw: {pw}")
            connection_dto = EmulationConnectionObservationState(
                conn=target_conn, credential=cr, root=root, service=constants.SSH.SERVICE_NAME,
                port=constants.SSH.DEFAULT_PORT, proxy=proxy_conn, ip=target_ip)
            return connection_dto
        except Exception as e:
            Logger.__call__().get_logger().warning(f"Custom connection setup failed: {str(e)}, {repr(e)}")
            return None

    @staticmethod
    def write_remote_file(conn, file_name: str, contents: str, write_mode: str = "w") -> None:
        """
        Utility function for writing contents to a file

        :param conn: the SSH connection to use for writing
        :param file_name: the file name
        :param contents: the contents of the file
        :param write_mode: the write mode
        :return: None
        """
        sftp_client = conn.open_sftp()
        remote_file = sftp_client.file(file_name, mode=write_mode)
        try:
            remote_file.write(contents)
        except Exception as e:
            Logger.__call__().get_logger().warning(f"Exception writing file: {str(e)}, {repr(e)}")
        finally:
            remote_file.close()

    @staticmethod
    def connect_admin(emulation_env_config: EmulationEnvConfig, ip: str,
                      create_producer: bool = False) -> None:
        """
        Connects the admin agent

        :param emulation_env_config: the configuration of the emulation to connect to
        :param ip: the ip of the container to connect to
        :param create_producer: boolean flag whether the producer thread to Kafka should be created also
        :return: None
        """
        emulation_env_config.agent_ip = ip
        emulation_env_config.connect(ip=ip, username=constants.CSLE_ADMIN.SSH_USER, pw=constants.CSLE_ADMIN.SSH_PW,
                                     create_producer=create_producer)

    @staticmethod
    def disconnect_admin(emulation_env_config: EmulationEnvConfig) -> None:
        """
        Disconnects the admin agent

        :param emulation_env_config: the configuration of the emulation to disconnect the admin of
        :return: None
        """
        emulation_env_config.close_all_connections()

    @staticmethod
    def execute_cmd_interactive_channel(cmd: str, channel) -> None:
        """
        Executes an action on the emulation using an interactive shell (non synchronous)
        :param a: action to execute
        :param env_config: environment config
        :param channel: the channel to use
        :return: None
        """
        channel.send(cmd + "\n")

    @staticmethod
    def read_result_interactive(emulation_env_config: EmulationEnvConfig, channel) -> str:
        """
        Reads the result of an action executed in interactive mode

        :param emulation_env_config: the emulation environment config
        :param the channel
        :return: the result
        """
        return EmulationUtil.read_result_interactive_channel(
            emulation_env_config=emulation_env_config,
            channel=channel)

    @staticmethod
    def read_result_interactive_channel(emulation_env_config: EmulationEnvConfig, channel) -> str:
        """
        Reads the result of an action executed in interactive mode
        :param emulation_env_config: the emulation environment config
        :param channel: the channel to use
        :return: the result
        """
        while not channel.recv_ready():
            time.sleep(constants.ENV_CONSTANTS.SHELL_READ_WAIT)
        output = channel.recv(constants.COMMON.LARGE_RECV_SIZE)
        output_str = output.decode("utf-8")
        output_str = constants.NMAP.SHELL_ESCAPE.sub("", output_str)
        return str(output_str)

    @staticmethod
    def is_emulation_defense_action_legal(defense_action_id: int, env_config: EmulationEnvConfig,
                                          env_state: EmulationEnvState) -> bool:
        """
        Checks if a given defense action is legal in the current state of the environment

        :param defense_action_id: the id of the action to check
        :param env_config: the environment config
        :param env_state: the environment state
        :param attacker_action: the id of the previous attack action
        :return: True if legal, else false
        """
        return True

    @staticmethod
    def is_emulation_attack_action_legal(attack_action_id: int, env_config: EmulationEnvConfig,
                                         env_state: EmulationEnvState) -> bool:
        """
        Checks if a given attack action is legal in the current state of the environment

        :param attack_action_id: the id of the action to check
        :param env_config: the environment config
        :param env_state: the environment state
        :return: True if legal, else false
        """
        if attack_action_id > len(env_state.attacker_action_config.actions) - 1:
            return False
        action = env_state.attacker_action_config.actions[attack_action_id]
        if env_state.attacker_obs_state is None:
            raise ValueError("Could not obtain EmulationAttackerObservationState")
        ip = env_state.attacker_obs_state.get_action_ips(action, emulation_env_config=env_config)

        logged_in_ips_str = EnvDynamicsUtil.logged_in_ips_str(emulation_env_config=env_config, s=env_state)
        if (action.id, action.index, logged_in_ips_str) in env_state.attacker_obs_state.actions_tried:
            return False

        # Recon on subnet is always possible
        if action.type == EmulationAttackerActionType.RECON and action.index == -1:
            return True

        # Recon on set of all found machines is always possible if there exists such machiens
        if action.type == EmulationAttackerActionType.RECON and action.index == -1 and len(
                env_state.attacker_obs_state.machines) > 0:
            return True

        # Optimal Stopping actions are always possible
        if action.type == EmulationAttackerActionType.STOP or action.type == EmulationAttackerActionType.CONTINUE:
            return True

        machine_discovered = False
        target_machine = None
        logged_in = False
        unscanned_filesystems = False
        untried_credentials = False
        root_login = False
        machine_root_login = False
        machine_logged_in = False
        uninstalled_tools = False
        machine_w_tools = False
        uninstalled_backdoor = False
        target_untried_credentials = False

        for m in env_state.attacker_obs_state.machines:
            if m.logged_in:
                logged_in = True
                if not m.filesystem_searched:
                    unscanned_filesystems = True
                if m.root:
                    root_login = True
                    if not m.tools_installed and not m.install_tools_tried:
                        uninstalled_tools = True
                    else:
                        machine_w_tools = True
                    if m.tools_installed and not m.backdoor_installed and not m.backdoor_tried:
                        uninstalled_backdoor = True
            if m.ips == ip:
                machine_discovered = True
                target_machine = m
                if m.logged_in:
                    machine_logged_in = True
                    if m.root:
                        machine_root_login = True
                if m.untried_credentials:
                    target_untried_credentials = m.untried_credentials
            # if m.shell_access and not m.logged_in:
            #     untried_credentials = True
            if m.untried_credentials:
                untried_credentials = m.untried_credentials

        if action.index == -1 or action.id == EmulationAttackerActionId.NETWORK_SERVICE_LOGIN:
            machine_discovered = True

        # Privilege escalation only legal if machine discovered and logged in and not root
        if (action.type == EmulationAttackerActionType.PRIVILEGE_ESCALATION and
                (not machine_discovered or not machine_logged_in or machine_root_login)):
            return False

        # Exploit only legal if we have not already compromised the node
        if action.type == EmulationAttackerActionType.EXPLOIT and machine_logged_in and root_login:
            return False

        # Shell-access Exploit only legal if we do not already have untried credentials
        if (action.type == EmulationAttackerActionType.EXPLOIT
                and action.action_outcome == EmulationAttackerActionOutcome.SHELL_ACCESS
                and target_untried_credentials):
            return False

        # Priv-Esc Exploit only legal if we are already logged in and do not have root
        if action.type == EmulationAttackerActionType.EXPLOIT \
                and action.action_outcome == EmulationAttackerActionOutcome.PRIVILEGE_ESCALATION_ROOT \
                and (not machine_logged_in or root_login):
            return False

        # If IP is discovered, then IP specific action without other prerequisites is legal
        if (machine_discovered and (action.type == EmulationAttackerActionType.RECON or
                                    action.type == EmulationAttackerActionType.EXPLOIT or
                                    action.type == EmulationAttackerActionType.PRIVILEGE_ESCALATION)):
            if action.index == -1 and target_machine is None:
                return True
            else:
                exploit_tried = env_state.attacker_obs_state.exploit_tried(a=action, m=target_machine)
            if exploit_tried:
                return False
            return True

        # If nothing new to scan, find-flag is illegal
        if action.id == EmulationAttackerActionId.FIND_FLAG and not unscanned_filesystems:
            return False

        # If nothing new to backdoor, install backdoor is illegal
        if action.id == EmulationAttackerActionId.SSH_BACKDOOR and not uninstalled_backdoor:
            return False

        # If no new credentials, login to service is illegal
        if action.id == EmulationAttackerActionId.NETWORK_SERVICE_LOGIN and not untried_credentials:
            return False

        # Pivot recon possible if logged in on pivot machine with tools installed
        if (machine_discovered and action.type == EmulationAttackerActionType.POST_EXPLOIT and
                logged_in and machine_w_tools):
            return True

        # If IP is discovered, and credentials are found and shell access, then post-exploit actions are legal
        if machine_discovered and action.type == EmulationAttackerActionType.POST_EXPLOIT \
                and ((target_machine is not None and target_machine.shell_access
                      and len(target_machine.shell_access_credentials) > 0)
                     or action.index == -1 or action.id == EmulationAttackerActionId.NETWORK_SERVICE_LOGIN):
            return True

        # Bash action not tied to specific IP only possible when having shell access and being logged in
        if action.id == EmulationAttackerActionId.FIND_FLAG and logged_in and unscanned_filesystems:
            return True

        # Bash action not tied to specific IP only possible when having shell access and being logged in and root
        if action.id == EmulationAttackerActionId.INSTALL_TOOLS and logged_in and root_login and uninstalled_tools:
            return True

        # Bash action not tied to specific IP only possible when having shell access and being logged in and root
        if action.id == EmulationAttackerActionId.SSH_BACKDOOR and logged_in and root_login \
                and machine_w_tools and uninstalled_backdoor:
            return True

        return False

    @staticmethod
    def check_pid(pid: int) -> bool:
        """
        Check if a given pid is running on the host

        :param pid: the pid to check
        :return: True if it is running otherwise false.
        """
        try:
            return bool(psutil.Process(pid).is_running())
        except Exception:
            return False

    @staticmethod
    def physical_ip_match(emulation_env_config: EmulationEnvConfig, ip, physical_host_ip: str) -> bool:
        """
        Utility method for checking if a container ip matches a physical ip

        :param emulation_env_config: the emulation config
        :param ip: the ip to check
        :param physical_host_ip:  the physical ip
        :return: True or False
        """
        if ip == emulation_env_config.kafka_config.container.docker_gw_bridge_ip and \
                emulation_env_config.kafka_config.container.physical_host_ip == physical_host_ip:
            return True
        if ip == emulation_env_config.elk_config.container.docker_gw_bridge_ip and \
                emulation_env_config.elk_config.container.physical_host_ip == physical_host_ip:
            return True
        if emulation_env_config.sdn_controller_config is not None and \
                ip == emulation_env_config.sdn_controller_config.container.docker_gw_bridge_ip and \
                emulation_env_config.sdn_controller_config.container.physical_host_ip == physical_host_ip:
            return True
        node_config = emulation_env_config.containers_config.get_container_from_ip(ip=ip)
        if node_config is not None and node_config.physical_host_ip == physical_host_ip:
            return True
        return False



from typing import List, Dict, Any
from csle_common.dao.simulation_config.player_config import PlayerConfig
from csle_base.json_serializable import JSONSerializable


class PlayersConfig(JSONSerializable):
    """
    A DTO representing the configuration of players in a simulation environment
    """

    def __init__(self, player_configs: List[PlayerConfig]):
        """
        Initializes the DTO

        :param player_configs: the list of configurations for each player
        """
        self.player_configs = player_configs

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation
        
        :return: a dict representation of the object
        """
        d = {}
        d["player_configs"] = list(map(lambda x: x.to_dict(), self.player_configs))
        return d

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "PlayersConfig":
        """
        Converts a dict representation of the object to a DTO

        :param d: the dict to convert
        :return: the created object
        """
        obj = PlayersConfig(
            player_configs=list(map(lambda x: PlayerConfig.from_dict(x), d["player_configs"]))
        )
        return obj

    def __str__(self):
        """
        :return: a string representation of the DTO
        """
        return f"players_configs: {list(map(lambda x: str(x), self.player_configs))}"

    @staticmethod
    def from_json_file(json_file_path: str) -> "PlayersConfig":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return PlayersConfig.from_dict(json.loads(json_str))



from typing import List, Dict, Union, Any
import numpy as np
from numpy.typing import NDArray
import torch
from stable_baselines3 import DQN
from csle_common.dao.training.policy import Policy
from csle_common.dao.training.agent_type import AgentType
from csle_common.dao.training.player_type import PlayerType
from csle_common.dao.simulation_config.state import State
from csle_common.dao.simulation_config.state_type import StateType
from csle_common.dao.simulation_config.action import Action
from csle_common.dao.training.experiment_config import ExperimentConfig
from csle_common.dao.training.policy_type import PolicyType
from csle_common.models.q_network import QNetwork
from csle_common.logging.log import Logger


class DQNPolicy(Policy):
    """
    A neural network policy learned with DQN
    """

    def __init__(self, model: Union[None, DQN, QNetwork], simulation_name: str, save_path: str,
                 player_type: PlayerType, states: List[State], actions: List[Action],
                 experiment_config: ExperimentConfig, avg_R: float):
        """
        Initializes the policy

        :param model: the DQN model
        :param simulation_name: the simulation name
        :param save_path: the path to save the model to
        :param states: list of states (required for computing stage policies)
        :param actions: list of actions
        :param experiment_config: the experiment configuration for training the policy
        :param avg_R: the average reward of the policy when evaluated in the simulation
        """
        super(DQNPolicy, self).__init__(agent_type=AgentType.DQN, player_type=player_type)
        self.model = model
        self.id = -1
        self.simulation_name = simulation_name
        self.save_path = save_path
        if self.model is None and save_path != "":
            try:
                self.model = DQN.load(path=self.save_path)
            except Exception as e:
                try:
                    self.model = QNetwork.load(path=self.save_path)
                except Exception as e2:
                    Logger.__call__().get_logger().warning(
                        f"There was an exception loading the model from path: {self.save_path}, "
                        f"exception: {str(e)}, {repr(e)}, {str(e2)}, {repr(e2)}")
        self.states = states
        self.actions = actions
        self.experiment_config = experiment_config
        self.avg_R = avg_R
        self.policy_type = PolicyType.DQN

    def action(self, o: List[float], deterministic: bool = True) -> NDArray[Any]:
        """
        Multi-threshold stopping policy

        :param o: the current observation
        :param deterministic: boolean flag indicating whether the action selection should be deterministic
        :return: the selected action
        """
        if self.model is None:
            raise ValueError("The model i None")
        if isinstance(self.model, DQN):
            a = self.model.predict(np.array(o), deterministic=deterministic)[0]
        elif isinstance(self.model, QNetwork):
            a = self.model.forward(x=torch.Tensor(o)).numpy()
        else:
            raise ValueError(f"Invalid model: {self.model}")
        return a

    def probability(self, o: List[float], a) -> int:
        """
        Multi-threshold stopping policy

        :param o: the current observation
        :return: the selected action
        """
        if self.model is None:
            raise ValueError("The model is None")
        if isinstance(self.model, DQN):
            actions = self.model.policy.forward(obs=torch.tensor(o).to(self.model.device))
            action = actions[0]
            if action == a:
                return 1
            else:
                return 0
        else:
            raise ValueError(f"This function is not supported for the current model: {self.model}")

    def to_dict(self) -> Dict[str, Any]:
        """
        :return: a dict representation of the policy
        """
        d: Dict[str, Any] = {}
        d["id"] = self.id
        d["simulation_name"] = self.simulation_name
        d["save_path"] = self.save_path
        d["states"] = list(map(lambda x: x.to_dict(), self.states))
        d["player_type"] = self.player_type
        d["actions"] = list(map(lambda x: x.to_dict(), self.actions))
        d["experiment_config"] = self.experiment_config.to_dict()
        d["agent_type"] = self.agent_type
        d["avg_R"] = self.avg_R
        d["policy_type"] = self.policy_type
        return d

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "DQNPolicy":
        """
        Converst a dict representation of the object to an instance

        :param d: the dict to convert
        :return: the created instance
        """
        obj = DQNPolicy(model=None, simulation_name=d["simulation_name"], save_path=d["save_path"],
                        states=list(map(lambda x: State.from_dict(x), d["states"])), player_type=d["player_type"],
                        actions=list(map(lambda x: Action.from_dict(x), d["actions"])),
                        experiment_config=ExperimentConfig.from_dict(d["experiment_config"]), avg_R=d["avg_R"])
        obj.id = d["id"]
        return obj

    def stage_policy(self, o: Union[List[int], List[float]]) -> List[List[float]]:
        """
        Gets the stage policy, i.e a |S|x|A| policy

        :param o: the latest observation
        :return: the |S|x|A| stage policy
        """
        b1 = o[1]
        l = int(o[0])
        if not self.player_type == PlayerType.ATTACKER:
            stage_policy = []
            for _ in self.states:
                stage_policy.append(self._get_attacker_stopping_dist(obs=o))
            return stage_policy
        else:
            stage_policy = []
            for s in self.states:
                if s.state_type != StateType.TERMINAL:
                    o = [l, b1, s.id]
                    stage_policy.append(self._get_attacker_stopping_dist(obs=o))
                else:
                    stage_policy.append([0.5, 0.5])
            return stage_policy

    def _get_attacker_stopping_dist(self, obs) -> List[float]:
        """
        Utility function for getting the stopping action distribution conditioned on a given observation

        :param obs: the observation to condition on
        :return: the conditional ation distribution
        """
        if self.model is None:
            raise ValueError("The model is None")
        obs = np.array([obs])
        if isinstance(self.model, DQN):
            actions = self.model.policy.forward(obs=torch.tensor(obs).to(self.model.device))
        else:
            raise ValueError(f"Stopping dist not supported for model: {self.model}")
        action = actions[0]
        if action == 1:
            stop_prob = 1
        else:
            stop_prob = 0
        return [1 - stop_prob, stop_prob]

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        return f"model: {self.model}, id: {self.id}, simulation_name: {self.simulation_name}, " \
               f"save path: {self.save_path}, states: {self.states}, experiment_config: {self.experiment_config}," \
               f"avg_R: {self.avg_R}, policy_type: {self.policy_type}"

    @staticmethod
    def from_json_file(json_file_path: str) -> "DQNPolicy":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return DQNPolicy.from_dict(json.loads(json_str))

    def copy(self) -> "DQNPolicy":
        """
        :return: a copy of the DTO
        """
        return self.from_dict(self.to_dict())



from typing import Union, List, Dict, Any, Optional
import numpy as np
from numpy.typing import NDArray
from csle_common.dao.training.agent_type import AgentType
from csle_common.dao.training.player_type import PlayerType
from csle_common.dao.training.policy import Policy
from csle_common.dao.simulation_config.action import Action
from csle_common.dao.training.policy_type import PolicyType


class TabularPolicy(Policy):
    """
    Object representing a tabular policy
    """

    def __init__(self, player_type: PlayerType, actions: List[Action], lookup_table: List[List[float]],
                 agent_type: AgentType, simulation_name: str, avg_R: float,
                 value_function: Optional[List[Any]] = None, q_table: Optional[List[Any]] = None) -> None:
        """
        Initializes the policy

        :param actions: list of actions
        :param player_type: the player type
        :param lookup_table: the lookup table that defines the policy
        :param value_function: the value function (optional)
        :param q_table: the Q-value function (optional)
        :param simulation_name: the name of the simulation
        :param avg_R: average reward obtained with the policy
        """
        super(TabularPolicy, self).__init__(agent_type=agent_type, player_type=player_type)
        self.actions = actions
        self.lookup_table = lookup_table
        self.value_function = value_function
        self.q_table = q_table
        self.simulation_name = simulation_name
        self.id = -1
        self.avg_R = avg_R
        self.policy_type = PolicyType.TABULAR

    def action(self, o: Union[int, float], deterministic: bool = True) -> Union[int, List[int], float, NDArray[Any]]:
        """
        Selects the next action

        :param o: the input observation
        :param deterministic: boolean flag indicating whether the action selection should be deterministic
        :return: the next action and its probability
        """
        return int(np.random.choice(np.arange(0, len(self.lookup_table[int(o)])), p=self.lookup_table[int(o)]))

    def probability(self, o: Union[int, float], a: int) -> float:
        """
        Calculates the probability of taking a given action for a given observation

        :param o: the input observation
        :param a: the action
        :return: p(a|o)
        """
        return self.lookup_table[int(o)][a]

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "TabularPolicy":
        """
        Converts a dict representation to an instance

        :param d: the dict to convert
        :return: the created instance
        """
        dto = TabularPolicy(actions=list(map(lambda x: Action.from_dict(x), d["actions"])),
                            player_type=d["player_type"], agent_type=d["agent_type"],
                            lookup_table=d["lookup_table"], value_function=d["value_function"],
                            simulation_name=d["simulation_name"], avg_R=d["avg_R"],
                            q_table=d["q_table"])
        if "id" in d:
            dto.id = d["id"]
        return dto

    def to_dict(self) -> Dict[str, Any]:
        """
        :return: A dict representation of the function
        """
        d: Dict[str, Any] = {}
        d["agent_type"] = self.agent_type
        d["player_type"] = self.player_type
        d["actions"] = list(map(lambda x: x.to_dict(), self.actions))
        d["lookup_table"] = self.lookup_table
        d["value_function"] = self.value_function
        d["simulation_name"] = self.simulation_name
        d["id"] = self.id
        d["avg_R"] = self.avg_R
        d["q_table"] = self.q_table
        d["policy_type"] = self.policy_type
        return d

    def stage_policy(self, o: Union[List[Union[int, float]], int, float]) -> List[List[float]]:
        """
        Gets the stage policy, i.e a |S|x|A| policy

        :param o: the latest observation
        :return: the |S|x|A| stage policy
        """
        return self.lookup_table

    def __str__(self) -> str:
        """
        :return: a string representation of the policy
        """
        return f"agent_type: {self.agent_type}, player_type: {self.player_type}, " \
               f"actions: {list(map(lambda x: str(x), self.actions))}, lookup_table: {self.lookup_table}, " \
               f"value_function: {self.value_function}, simulation_name: {self.simulation_name}, id: {self.id}, " \
               f"avg_R: {self.avg_R}, q_table: {self.q_table}, policy_type: {self.policy_type}"

    @staticmethod
    def from_json_file(json_file_path: str) -> "TabularPolicy":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return TabularPolicy.from_dict(json.loads(json_str))

    def copy(self) -> "TabularPolicy":
        """
        :return: a copy of the DTO
        """
        return self.from_dict(self.to_dict())



"""
Type of defense actions
"""
from enum import IntEnum


class EmulationDefenderActionType(IntEnum):
    """
    Enum representing the different defense types in the network.
    """
    STOP = 0
    CONTINUE = 1
    ADD_DEFENSIVE_MECHANISM = 2



from typing import Tuple, List
import socket
import numpy as np
import numpy.typing as npt


class GeneralUtil:
    """
    Class with general utility functions
    """

    @staticmethod
    def get_host_ip() -> str:
        """
        Utility method for getting the ip of the host

        :return: the ip of the host
        """
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        s.connect(("8.8.8.8", 80))
        return str(s.getsockname()[0])

    @staticmethod
    def replace_first_octet_of_ip(ip: str, ip_first_octet: int) -> str:
        """
        Utility function for changing the first octet in an IP address

        :param ip: the IP to modify
        :param ip_first_octet: the first octet to insert
        :return: the new IP
        """
        index_of_first_octet_end = ip.find(".")
        return str(ip_first_octet) + ip[index_of_first_octet_end:]

    @staticmethod
    def replace_first_octet_of_ip_tuple(tuple_of_ips: Tuple[str, str], ip_first_octet: int) -> Tuple[str, str]:
        """
        Utility function for changing the first octet in an IP address

        :param ip: the IP to modify
        :param ip_first_octet: the first octet to insert
        :return: the new IP
        """
        index_of_first_octet_end = tuple_of_ips[0].find(".")
        first_ip = str(ip_first_octet) + tuple_of_ips[0][index_of_first_octet_end:]
        index_of_first_octet_end = tuple_of_ips[1].find(".")
        second_ip = str(ip_first_octet) + tuple_of_ips[1][index_of_first_octet_end:]
        return (first_ip, second_ip)

    @staticmethod
    def get_latest_table_id(cur, table_name: str) -> int:
        """
        Gets the next ID for a table with a serial column primary key

        :param cur: the postgres connection cursor
        :param table_name: the table name
        :return: the next id
        """
        cur.execute(f"SELECT id FROM {table_name}")
        id = 1
        ids = cur.fetchall()
        if len(ids) > 0:
            id = max(list(map(lambda x: x[0], ids))) + 1
        return id

    @staticmethod
    def list_to_tuple(L):
        """
        Converts a nested list to a tuple
        """
        if isinstance(L, list):
            return tuple(GeneralUtil.list_to_tuple(item) for item in L)  # Recursively convert lists to tuples
        return L

    @staticmethod
    def one_hot_encode_integer(value: int, max_value: int) -> npt.NDArray[np.int32]:
        """
        One-hot encodes an integer

        :param value: the interger value
        :param max_value: the maximum value
        :return: the one-hot encoded vector
        """
        if not (0 <= value <= max_value):
            raise ValueError(f"Value is: {value}, Value must be between 0 and {max_value} (inclusive).")
        one_hot_vector = np.zeros(max_value + 1, dtype=np.int32)
        one_hot_vector[value] = 1
        return one_hot_vector

    @staticmethod
    def one_hot_encode_vector(vector: List[int], max_value: int) -> npt.NDArray[np.int32]:
        """
        One-hot encodes a vector

        :param vector: the vector to one-hot encode
        :param max_value: the maximum value in each entry of the vector
        :return: the one-hot encoded vector
        """
        np_vector = np.array(vector)

        # Validate input values
        if not np.all((0 <= np_vector) & (np_vector <= max_value)):
            raise ValueError(f"Vector is: {np_vector}. Input vector can only contain values between 0 and {max_value}.")

        # Create one-hot encoding dynamically based on max_value
        one_hot = np.eye(max_value + 1)[np_vector]  # Identity matrix of size (max_value + 1)

        return np.array(one_hot.flatten())



from typing import List
import threading
import time
from confluent_kafka import Consumer, KafkaError, KafkaException
import csle_collector.constants.constants as collector_constants
from csle_collector.client_manager.client_population_metrics import ClientPopulationMetrics
from csle_common.logging.log import Logger


class ClientPopulationConsumerThread(threading.Thread):
    """
    Thread that polls the Kafka log to get the latest status of the client population
    """

    def __init__(self, kafka_server_ip: str, kafka_port: int,
                 client_population_metrics: ClientPopulationMetrics, auto_offset_reset: str = "latest") -> None:
        """
        Initialzes the thread

        :param kafka_server_ip: the ip of the kafka server
        :param kafka_port: the port of the kafka server
        :param client_population_metrics: the client population metrics to update
        :param auto_offset_reset: the offset for kafka to start reading from
        """
        threading.Thread.__init__(self)
        self.running = True
        self.kafka_server_ip = kafka_server_ip
        self.kafka_port = kafka_port
        self.client_population_metrics = client_population_metrics
        self.client_population_metrics_list: List[ClientPopulationMetrics] = []
        self.ts = time.time()
        self.auto_offset_reset = auto_offset_reset
        self.kafka_conf = {
            collector_constants.KAFKA.BOOTSTRAP_SERVERS_PROPERTY: f"{self.kafka_server_ip}:{self.kafka_port}",
            collector_constants.KAFKA.GROUP_ID_PROPERTY: f"client_population_consumer_thread_{self.ts}",
            collector_constants.KAFKA.AUTO_OFFSET_RESET_PROPERTY: auto_offset_reset}
        self.consumer = Consumer(**self.kafka_conf)
        self.consumer.subscribe([collector_constants.KAFKA_CONFIG.CLIENT_POPULATION_TOPIC_NAME])

    def run(self) -> None:
        """
        Run thread

        :return: None
        """
        while self.running:
            msg = self.consumer.poll(timeout=5.0)
            if msg is not None:
                if msg.error():
                    if msg.error().code() == KafkaError._PARTITION_EOF:
                        Logger.__call__().get_logger.warning(
                            f"reached end of partition: {msg.topic(), msg.partition(), msg.offset()}")
                    elif msg.error():
                        raise KafkaException(msg.error())
                else:
                    self.client_population_metrics.update_with_kafka_record(record=msg.value().decode())
                    self.client_population_metrics_list.append(self.client_population_metrics.copy())
        self.consumer.close()

    def get_average_client_population_metrics(self) -> ClientPopulationMetrics:
        """
        :return: average of the list of client population metrics
        """
        if len(self.client_population_metrics_list) == 0:
            return self.client_population_metrics.copy()
        if len(self.client_population_metrics_list) == 1:
            return self.client_population_metrics_list[0].copy()
        avg_client_population_metrics = ClientPopulationMetrics(num_clients=0, rate=0)

        for i in range(len(self.client_population_metrics_list)):
            avg_client_population_metrics.num_clients = (avg_client_population_metrics.num_clients +
                                                         self.client_population_metrics_list[i].num_clients)
            avg_client_population_metrics.rate = (avg_client_population_metrics.rate +
                                                  self.client_population_metrics_list[i].rate)

        avg_client_population_metrics.num_clients = int(round(
            avg_client_population_metrics.num_clients / len(self.client_population_metrics_list)))
        avg_client_population_metrics.rate = float(round(
            avg_client_population_metrics.rate / len(self.client_population_metrics_list)))
        return avg_client_population_metrics



import csle_common.constants.constants as constants
from csle_common.util.management_util import ManagementUtil
from unittest.mock import patch


class TestManagementUtilSuite:
    """
    Test suite for management util
    """

    @patch("csle_common.metastore.metastore_facade.MetastoreFacade.list_management_users")
    @patch("bcrypt.gensalt")
    @patch("bcrypt.hashpw")
    @patch("csle_common.metastore.metastore_facade.MetastoreFacade.save_management_user")
    def test_create_default_management_admin_account(self, mock_save_management_user, mock_hashpw, mock_gensalt,
                                                     mock_list_management_users) -> None:
        """
        Test the method that creates the default management admin account

        :param mock_save_management_user: mock_save_management_user
        :param mock_hashpw: mock_hashpw
        :param mock_gensalt: mock_gensalt
        :param mock_list_management_users: mock_list_management_users
        :return: None
        """
        mock_list_management_users.return_value = []
        mock_salt = b"salt"
        mock_gensalt.return_value = mock_salt
        mock_hash = b"hashed_password"
        mock_hashpw.return_value = mock_hash
        constants.CSLE_ADMIN.MANAGEMENT_USER = "admin"
        constants.CSLE_ADMIN.MANAGEMENT_PW = "password"
        constants.CSLE_ADMIN.MANAGEMENT_FIRST_NAME = "first"
        constants.CSLE_ADMIN.MANAGEMENT_LAST_NAME = "last"
        constants.CSLE_ADMIN.MANAGEMENT_ORGANIZATION = "organization"
        constants.CSLE_ADMIN.MANAGEMENT_EMAIL = "admin@email.com"
        ManagementUtil.create_default_management_admin_account()
        mock_list_management_users.assert_called_once()
        mock_gensalt.assert_called_once()
        mock_hashpw.assert_called_once_with(constants.CSLE_ADMIN.MANAGEMENT_PW.encode("utf-8"), mock_salt)
        mock_save_management_user.assert_called_once()

    @patch("csle_common.metastore.metastore_facade.MetastoreFacade.list_management_users")
    @patch("bcrypt.gensalt")
    @patch("bcrypt.hashpw")
    @patch("csle_common.metastore.metastore_facade.MetastoreFacade.save_management_user")
    def test_create_default_management_guest_account(self, mock_save_management_user, mock_hashpw, mock_gensalt,
                                                     mock_list_management_users) -> None:
        """
        Test the method that creates the default management guest account

        :param mock_save_management_user: mock_save_management_user
        :param mock_hashpw: mock_hashpw
        :param mock_gensalt: mock_gensalt
        :param mock_list_management_users: mock_list_management_users
        :return: None
        """
        mock_list_management_users.return_value = []
        mock_salt = b"salt"
        mock_gensalt.return_value = mock_salt
        mock_hash = b"hashed_password"
        mock_hashpw.return_value = mock_hash
        constants.CSLE_GUEST.MANAGEMENT_USER = "user"
        constants.CSLE_GUEST.MANAGEMENT_PW = "password"
        constants.CSLE_GUEST.MANAGEMENT_FIRST_NAME = "guest_first"
        constants.CSLE_GUEST.MANAGEMENT_LAST_NAME = "guest_last"
        constants.CSLE_GUEST.MANAGEMENT_ORGANIZATION = "guest_organization"
        constants.CSLE_GUEST.MANAGEMENT_EMAIL = "guest@email.com"
        ManagementUtil.create_default_management_guest_account()
        mock_list_management_users.assert_called_once()
        mock_gensalt.assert_called_once()
        mock_hashpw.assert_called_once_with(constants.CSLE_GUEST.MANAGEMENT_PW.encode("utf-8"), mock_salt)
        mock_save_management_user.assert_called_once()
